import math, re

from coldtype.geometry.geometrical import Geometrical
from coldtype.geometry.point import Point
from coldtype.geometry.line import Line
from coldtype.geometry.edge import Edge, txt_to_edge
from coldtype.interpolation import norm
from coldtype.geometry.primitives import *

try:
    from fontTools.misc.transform import Transform
except ImportError:
    Transform = None


COMMON_PAPER_SIZES = {
    'letter': (612, 792),
    'tabloid': (792, 1224),
    'ledger': (1224, 792),
    'legal': (612, 1008),
    'a0': (2384, 3371),
    'a1': (1685, 2384),
    'a2': (1190, 1684),
    'a3': (842, 1190),
    'a4': (595, 842),
    'a4Small': (595, 842),
    'a5': (420, 595),
    'b4': (729, 1032),
    'b5': (516, 729),
    'folio': (612, 936),
    'quarto': (610, 780),
    '10x14': (720, 1008),
}

for key, (w, h) in list(COMMON_PAPER_SIZES.items()):
    COMMON_PAPER_SIZES["%s-landscape" % key] = (h, w)


def align(b, rect, x=Edge.CenterX, y=Edge.CenterY):
    x = txt_to_edge(x)
    y = txt_to_edge(y)
    
    xoff = 0
    if x != None:
        if x == Edge.CenterX:
            xoff = -b.x + rect.x + rect.w/2 - b.w/2
        elif x == Edge.MinX:
            xoff = -(b.x-rect.x)
        elif x == Edge.MaxX:
            xoff = -b.x + rect.x + rect.w - b.w
    
    yoff = 0
    if y != None:
        if y == Edge.CenterY:
            yoff = -b.y + rect.y + rect.h/2 - b.h/2
        elif y == Edge.MaxY:
            yoff = (rect.y + rect.h) - (b.h + b.y)
        elif y == Edge.MinY:
            yoff = -(b.y-rect.y)
    
    #diff = rect.w - b.w
    return (xoff, yoff)


class Rect(Geometrical):
    """
    Representation of a rectangle as (x, y, w, h), indexable
    
    Constructor handles multiple formats, including:
    
    * ``x, y, w, h``
    * ``[x, y, w, h]``
    * ``w, h`` (x and y default to 0, 0)

    ``Rect`` objects can be splat'd where lists are expected as individual arguments (as in drawBot), i.e. ``rect(*my_rect)``, or can be passed directly to functions expected a list representation of a rectangle.
    """

    def FromCenter(center, w, h=None):
        """Create a rect given a center point and a width and height (optional, height will default to width if not specified")"""
        x, y = center
        if not h:
            h = w
        return Rect((x - w/2, y - h/2, w, h))
    
    def Inches(w, h, dpi=72.0):
        return Rect(w*dpi, h*dpi)

    def __init__(self, *rect):
        if isinstance(rect[0], str):
            x, y = 0, 0
            w, h = COMMON_PAPER_SIZES[rect[0].lower()]
        elif isinstance(rect[0], int) or isinstance(rect[0], float):
            try:
                x, y, w, h = rect
            except:
                w, h = rect
                x, y = 0, 0
        else:
            try:
                x, y, w, h = rect[0]
            except:
                w, h = rect[0]
                x, y = 0, 0
        self.x = x
        self.y = y
        self.w = w
        self.h = h
    
    def origin(self):
        """``(x, y)`` as tuple"""
        return self.x, self.y

    def from_obj(obj, w=None, h=None):
        r = Rect((0, 0, 0, 0))
        try:
            r.x = obj.x
            r.y = obj.y
        except:
            pass
        try:
            r.w = obj.w
            r.h = obj.h
        except:
            pass
        if w:
            r.w = w
            r.x -= w/2
        if h:
            r.h = h
            r.y -= h/2
        return r

    def FromExtents(extents):
        nw, ne, se, sw = extents
        return Rect(sw[0], sw[1], abs(ne[0] - sw[0]), abs(ne[1] - sw[1]))
    
    def noop(self, *args, **kwargs):
        return self

    def FromMnMnMxMx(extents):
        """Create a rectangle from ``xmin, ymin, xmax, ymax``"""
        xmin, ymin, xmax, ymax = extents
        return Rect(xmin, ymin, xmax - xmin, ymax - ymin)

    def FromPoints(*points):
        xmin, ymin, xmax, ymax = None, None, None, None
        for p in points:
            if xmin is None or p[0] < xmin:
                xmin = p[0]
            if ymin is None or p[1] < ymin:
                ymin = p[1]
            if xmax is None or p[0] > xmax:
                xmax = p[0]
            if ymax is None or p[1] > ymax:
                ymax = p[1]
        return Rect.FromMnMnMxMx([xmin, ymin, xmax, ymax])

    def mnmnmxmx(self):
        """Return extents of rectangle as list"""
        return (self.x, self.y, self.x + self.w, self.y + self.h)

    def __getitem__(self, key):
        return self.rect()[key]

    def __repr__(self):
        return "Rect(" + str(self.rect()).replace(" ", "") + ")"
    
    def __eq__(self, r):
        try:
            return all([self.x == r.x, self.y == r.y, self.w == r.w, self.h == r.h])
        except:
            return False
    
    __hash__ = object.__hash__

    def rect(self):
        """x,y,w,h in list"""
        return [self.x, self.y, self.w, self.h]
    
    def round(self):
        """round the values in the rectangle to the nearest integer"""
        return Rect([int(round(n)) for n in self])

    def xy(self):
        """equivalent to origin"""
        return [self.x, self.y]

    def wh(self):
        """the width and height as a tuple"""
        return [self.w, self.h]
    
    @property
    def mnx(self):
        return self.x
    
    @property
    def mny(self):
        return self.y

    @property
    def mxx(self):
        return self.x + self.w
    
    @property
    def mxy(self):
        return self.y + self.h
    
    @property
    def mdx(self):
        return self.point("C").x
    
    @property
    def mdy(self):
        return self.point("C").y

    def square(self):
        """take a square from the center of this rect"""
        return Rect(centered_square(self.rect()))
    
    def align(self, rect, x=Edge.CenterX, y=Edge.CenterY):
        return self.offset(*align(self, rect, x, y))
    
    def ipos(self, pt, defaults=(0.5, 0.5), clamp=True):
        """
        Get scaled 0-1 bounded (optional) value
        from a point in a rectangle
        """
        if not pt:
            return defaults
        sx = ((pt.x - self.x) / self.w)
        sy = ((pt.y - self.y) / self.h)
        if clamp:
            sx = min(1, max(0, sx))
            sy = min(1, max(0, sy))
        return sx, sy

    def divide(self, amount, edge, forcePixel=False):
        """
        **Dividing**

        Derived from the behavior of the classic Cocoa function CGRectDivide, which takes a rectangle and breaks it into two pieces, based on a pixel amount and an edge.

        A quick example: assume you have a rectangle, ``r``, defined as such:

        ``r = Rect(0, 0, 300, 100)``
        
        If you want to break that into a left-hand rectangle that’s 100 pixels wide and a right-hand rectangle that’s 200 pixels wide, you could either say:
        
        ``left, right = r.divide(100, "mnx")``
        
        `or you could say`
        
        ``right, left = r.divide(200, "mxx")``

        where ``mxx`` is the rightmost edge, and ``mnx`` is the leftmost edge.

        **Centering**

        A special use-case is if you want to break a rectangle into `three` rectangles, based on the center "edge", you can do something like this:

        ``left, center, right = r.divide(200, "mdx")``

        This will result in three rectangles, always left-to-right, where
        left is 50px wide, then center is 200px wide, then right is also 50px wide — anything not in the center will be evenly distributed between left and right, or top-and-bottom in the case of a Y edge.
        """
        edge = txt_to_edge(edge)
        if edge == Edge.CenterX or edge == Edge.CenterY:
            a, b, c = divide(self.rect(), amount, edge, forcePixel=forcePixel)
            return Rect(a), Rect(b), Rect(c)
        else:
            a, b = divide(self.rect(), amount, edge, forcePixel=forcePixel)
            return Rect(a), Rect(b)

    def subdivide(self, amount, edge):
        """
        Like ``divide``, but here you specify the number of equal pieces you want (like columns or rows), and then what edge to start at, i.e.
        
        .. code:: python
            
            r = Rect(0, 0, 500, 100)
            r.subdivide(5, "mxx")
            => [Rect([400.0, 0, 100.0, 100]), Rect([300.0, 0, 100.0, 100]), Rect([200.0, 0, 100.0, 100]), Rect([100.0, 0, 100.0, 100]), Rect([0, 0, 100.0, 100])]
        
        will get you five 100-px wide rectangles, right-to-left

        (N.B. Does not support center edges, as that makes no sense)
        """
        edge = txt_to_edge(edge)
        return [Rect(x) for x in subdivide(self.rect(), amount, edge)]
    
    def subdivide_with_leading(self, count, leading, edge, forcePixel=True):
        """
        Same as `subdivide`, but inserts leading between each subdivision
        """
        return self.subdivide_with_leadings(count, [leading]*(count-1), edge, forcePixel)

    def subdivide_with_leadings(self, count, leadings, edge, forcePixel=True):
        """
        Same as `subdivide_with_leadings`, but inserts leading between each subdivision, indexing the size of the leading from a list of leadings
        """
        edge = txt_to_edge(edge)
        leadings = leadings + [0]
        full = self.w if edge == Edge.MinX or edge == Edge.MaxX else self.h
        unit = (full - sum(leadings)) / count
        amounts = [val for pair in zip([unit] * count, leadings) for val in pair][:-1]
        return [Rect(x) for x in subdivide(self.rect(), amounts, edge, forcePixel=forcePixel)][::2]

    def transform(self, t):
        pts = ["NW", "NE", "SE", "SW"]
        x1, x2, x3, x4 = [t.transformPoint(self.point(pt)) for pt in pts]
        return Rect.FromExtents([x1, x2, x3, x4])

    def rotate(self, degrees, point=None):
        if Transform:
            t = Transform()
            if not point:
                point = self.point("C")
            t = t.translate(point.x, point.y)
            t = t.rotate(math.radians(degrees))
            t = t.translate(-point.x, -point.y)
            return self.transform(t)
        else:
            raise Exception("fontTools not installed")

    def scale(self, s, x_edge=Edge.MinX, y_edge=Edge.MinY):
        return Rect(scale(self.rect(), s, x_edge, y_edge))
        #x_edge = txt_to_edge(x_edge)
        #y_edge = txt_to_edge(y_edge)
        #sx = self.w * s
        #sy = self.h * s
        #return self.take(sx, x_edge, forcePixel=True).take(sy, y_edge, forcePixel=True)

    def union(self, otherRect):
        return Rect.FromMnMnMxMx(unionRect(self.mnmnmxmx(), otherRect.mnmnmxmx()))
    
    def intersection(self, otherRect):
        return Rect.FromMnMnMxMx(sectRect(self.mnmnmxmx(), otherRect.mnmnmxmx())[1])
    
    sect = intersection

    def take(self, amount, edge, forcePixel=False):
        """
        Like `divide`, but here it just returns the "first" rect from a divide call, not all the resulting pieces, i.e. you can "take" 200px from the center of a rectangle by doing this ``Rect(0, 0, 300, 100).take(200, "mdx")`` which will result in ``Rect([50, 0, 200, 100])``
        """
        edge = txt_to_edge(edge)
        return Rect(take(self.rect(), amount, edge, forcePixel=forcePixel))

    def takeOpposite(self, amount, edge, forcePixel=False):
        edge = txt_to_edge(edge)
        return self.divide(amount, edge, forcePixel=forcePixel)[1]

    def subtract(self, amount, edge):
        """
        The opposite of ``take``, this will remove and not return a piece of the given amount from the given edge.
        
        Let's say you have a 100px-wide square and you want to drop 10px from the right-hand side, you would do:

        ``Rect(100, 100).subtract(10, Edge.MaxX)``, which leaves you with ``Rect([0, 0, 90, 100])``
        """
        edge = txt_to_edge(edge)
        return Rect(subtract(self.rect(), amount, edge))

    def expand(self, amount, edge):
        edges = None
        if edge == "NW":
            edges = ["mxy", "mnx"]
        elif edge == "NE":
            edges = ["mxy", "mxx"]
        elif edge == "SE":
            edges = ["mny", "mxx"]
        elif edge == "SW":
            edges = ["mny", "mnx"]
        
        if edges:
            return self.expand(amount, edges[0]).expand(amount, edges[1])
        edge = txt_to_edge(edge)
        return Rect(expand(self.rect(), amount, edge))
    
    add = expand

    def inset(self, dx, dy=None):
        """
        Creates padding in the amount of dx and dy. Also does expansion with negative values, or both at once
        """
        if dy == None:
            dy = dx
        return Rect(inset(self.rect(), dx, dy))
    
    def inset_x(self, dx):
        return self.inset(dx, 0)
    
    def inset_y(self, dy):
        return self.inset(0, dy)

    def offset(self, dx, dy=None):
        if dy == None:
            dy = dx
        return Rect(offset(self.rect(), dx, dy))
    
    def offset_x(self, dx):
        return self.offset(dx, 0)
    
    def offset_y(self, dy):
        return self.offset(0, dy)
    
    o = offset

    def zero(self):
        return Rect((0, 0, self.w, self.h))

    def __add__(self, another_rect):
        return Rect(add(self, another_rect))

    def grid(self, columns=2, rows=2):
        """Construct a grid"""
        xs = [row.subdivide(columns, Edge.MinX) for row in self.subdivide(rows, Edge.MaxY)]
        return [item for sublist in xs for item in sublist]

    def pieces(self, amount, edge):
        edge = txt_to_edge(edge)
        return [Rect(x) for x in pieces(self.rect(), amount, edge)]

    def edge(self, edge):
        edge = txt_to_edge(edge)
        return Line(*edgepoints(self.rect(), edge))

    def center(self):
        return Point(centerpoint(self.rect()))

    def flip(self, h):
        return Rect([self.x, h - self.h - self.y, self.w, self.h])

    def cardinals(self):
        return self.point("N"), self.point("E"), self.point("S"), self.point("W")

    def intercardinals(self):
        return self.point("NE"), self.point("SE"), self.point("SW"), self.point("NW")
    
    def aspect(self):
        return self.h / self.w
    
    def fit(self, other):
        sx, sy, sw, sh = self
        ox, oy, ow, oh = other
        if ow > oh:
            fw = sw
            fh = sw * other.aspect()
        else:
            fh = sh
            sw = sh * 1/other.aspect()
        #print(sh, fw, fh, other.aspect())
        return self.take(fh, "mdy")
    
    def avg(self):
        pts = self.cardinals()
        return Point(
            sum([p.x for p in pts])/4,
            sum([p.y for p in pts])/4)

    def point(self, eh, ev=Edge.MinX):
        """
        Get a ``Point`` at a given compass direction, chosen from
        
        * C
        * W
        * NW
        * N
        * NE
        * E
        * SE
        * S
        * SW
        """

        ev = txt_to_edge(ev)
        pc = Edge.PairFromCompass(eh)
        if pc:
            return self.point(*pc)
        else:
            px = self.x
            py = self.y

            if eh == Edge.MaxX:
                px = self.x + self.w
            elif eh == Edge.CenterX:
                px = self.x + self.w/2

            if ev == Edge.MaxY:
                py = self.y + self.h
            if ev == Edge.CenterY:
                py = self.y + self.h/2

            return Point((px, py))
    
    p = point

    @property
    def pne(self): return self.point("NE")

    @property
    def pe(self): return self.point("E")

    @property
    def ee(self): return self.edge("mxx")

    @property
    def pse(self): return self.point("SE")

    @property
    def ps(self): return self.point("S")

    @property
    def es(self): return self.edge("mny")

    @property
    def psw(self): return self.point("SW")

    @property
    def pw(self): return self.point("W")

    @property
    def ew(self): return self.edge("mnx")

    @property
    def pnw(self): return self.point("NW")

    @property
    def pn(self): return self.point("N")

    @property
    def en(self): return self.edge("mxy")

    @property
    def pc(self): return self.point("C")

    @property
    def ecx(self): return self.edge("mdx")

    @property
    def ecy(self): return self.edge("mdy")

    def intersects(self, other):
        return not (self.point("NE").x < other.point("SW").x or self.point("SW").x > other.point("NE").x or self.point("NE").y < other.point("SW").y or self.point("SW").y > other.point("NE").y)
    
    def maxima(self, n, edge):
        e = txt_to_edge(edge)
        if e == Edge.MinX:
            return self.setmnx(n)
        elif e == Edge.MaxX:
            return self.setmxx(n)
        elif e == Edge.CenterX:
            return self.setmdx(n)
        elif e == Edge.MinY:
            return self.setmny(n)
        elif e == Edge.MaxY:
            return self.setmxy(n)
        elif e == Edge.CenterY:
            return self.setmdy(n)
        else:
            raise Exception("HELLO")
    
    def setmnx(self, x):
        mnx, mny, mxx, mxy = self.mnmnmxmx()
        return Rect.FromMnMnMxMx([x, mny, mxx, mxy])
    
    def __mul__(self, other):
        return self.setmnx(other)
    
    def setlmnx(self, x):
        if x > self.mnx:
            return self.setmnx(x)
        return self
    
    def setmny(self, y):
        mnx, mny, mxx, mxy = self.mnmnmxmx()
        return Rect.FromMnMnMxMx([mnx, y, mxx, mxy])
    
    def __matmul__(self, other):
        return self.setmny(other)
    
    def setlmny(self, y):
        if y > self.mny:
            return self.setmny(y)
        return self
    
    def setmxx(self, x):
        mnx, mny, mxx, mxy = self.mnmnmxmx()
        return Rect.FromMnMnMxMx([mnx, mny, x, mxy])
    
    def setlmxx(self, x):
        if x < self.mxx:
            return self.setmxx(x)
        return self
    
    def setmxy(self, y):
        mnx, mny, mxx, mxy = self.mnmnmxmx()
        return Rect.FromMnMnMxMx([mnx, mny, mxx, y])
    
    def setlmxy(self, y):
        if y < self.mxy:
            return self.setmxy(y)
        return self
    
    def setmdx(self, x):
        c = self.point("C")
        return Rect.FromCenter(Point([x, c.y]), self.w, self.h)
    
    def setmn(self, mn):
        return self.setmnx(mn.x).setmny(mn.y)

    def setmx(self, mx):
        return self.setmxx(mx.x).setmxy(mx.y)
    
    def setw(self, w):
        return Rect(self.x, self.y, w, self.h)
    
    def seth(self, h):
        return Rect(self.x, self.y, self.w, h)
    
    def parse_line(self, d, line):
        parts = re.split(r"\s|ƒ|,", line)
        reified = []
        for p in parts:
            if p == "auto" or p == "a":
                reified.append("auto")
            elif "%" in p:
                reified.append(float(p.replace("%", ""))/100 * d)
            else:
                fp = float(p)
                if fp > 1:
                    reified.append(fp)
                else:
                    reified.append(fp*d)
        remaining = d - sum([0 if r == "auto" else r for r in reified])
        #if not float(remaining).is_integer():
        #    raise Exception("floating parse")
        auto_count = reified.count("auto")
        if auto_count > 0:
            auto_d = remaining / auto_count
            auto_ds = [auto_d] * auto_count
            if not auto_d.is_integer():
                auto_d_floor = math.floor(auto_d)
                leftover = remaining - auto_d_floor * auto_count
                for aidx, ad in enumerate(auto_ds):
                    if leftover > 0:
                        auto_ds[aidx] = auto_d_floor + 1
                        leftover -= 1
                    else:
                        auto_ds[aidx] = auto_d_floor
                #print(auto_ds)
                #print("NO", auto_d - int(auto_d))
        res = []
        for r in reified:
            if r == "auto":
                res.append(auto_ds.pop())
            else:
                res.append(r)
        return res
        #return [auto_d if r == "auto" else r for r in reified]
    
    def __floordiv__(self, other):
        return self.offset(0, other)
    
    def __truediv__(self, other):
        return self.offset(other, 0)
    
    def symbol_to_edge(self, idx, symbol):
        d = ["x", "y"][idx]
        if symbol == "-":
            return "mn" + d
        elif symbol == "+":
            return "mx" + d
        elif symbol == "=":
            return "md" + d
    
    def sign_to_dim(self, sign):
        xy = "x"
        if isinstance(sign, complex):
            xy = "y"
            sign = sign.imag
        return xy, sign
    
    def sign_to_edge(self, sign):
        xy, sign = self.sign_to_dim(sign)
        if sign == 0:
            return "md" + xy
        elif sign < 0:
            return "mn" + xy
        else:
            return "mx" + xy
    
    def t(self, sign, n):
        return self.take(n, self.sign_to_edge(sign))
    
    def s(self, sign, n):
        return self.subtract(n, self.sign_to_edge(sign))
    
    def i(self, sign, n):
        xy, _ = self.sign_to_dim(sign)
        return self.inset(n if xy == "x" else 0, n if xy == "y" else 0)
    
    def columns(self, *args):
        r = self
        _xs = " ".join([str(s) for s in args])
        ws = self.parse_line(r.w, _xs)
        rs = []
        for w in ws:
            _r, r = r.divide(w, "mnx")
            rs.append(_r)
        return rs
    
    def rows(self, *args):
        r = self
        _xs = " ".join([str(s) for s in args])
        ws = self.parse_line(r.h, _xs)
        rs = []
        for w in ws:
            _r, r = r.divide(w, "mxy")
            rs.append(_r)
        return rs
    
    def to_pen(self):
        from coldtype.pens.draftingpen import DraftingPen
        return DraftingPen(self)


from pathlib import Path
from defcon import Font as DefconFont
from coldtype.text.reader import normalize_font_path, StyledString
from coldtype.pens.datpen import DATPens
from coldtype.interpolation import norm, interp_dict, lerp, loopidx
from random import Random


def sibling(root, file):
    return Path(root).parent.joinpath(file)

def raw_ufo(path):
    return DefconFont(normalize_font_path(path))

def ßhide(el):
    return None

def ßshow(el):
    return el

def cycle_idx(arr, idx):
    if idx < 0:
        return len(arr) - 1
    elif idx >= len(arr):
        return 0
    else:
        return idx

def random_series(start=0, end=1, seed=0, count=5000):
    rnd = Random()
    rnd.seed(seed)
    rnds = []
    for x in range(count):
        rnds.append(start+rnd.random()*(end-start))
    return rnds

def show_points(pen, style, offcurves=True, filter=lambda i: True):
    pt_labels = DATPens()
    if offcurves:
        def labeller(idx, x, y):
            if filter(idx):
                pt_labels.append(StyledString(str(idx), style).pen().translate(x, y))
        pen.map_points(labeller)
    else:
        for idx, (m, pts) in enumerate(pen.value):
            if len(pts) > 0 and filter(idx):
                pt_labels += StyledString(str(idx), style).pen().translate(*pts[-1])
    return pt_labels

_by_uni = None
_by_glyph = None
_class_lookup = None

def _populate_glyphs_unis():
    global _by_uni
    global _by_glyph
    global _class_lookup
    _by_uni = {}
    _by_glyph = {}
    _class_lookup = {}

    #try:
    if True:
        lines = (Path(__file__).parent.parent / "assets/glyphNamesToUnicode.txt").read_text().split("\n")

        for l in lines:
            if l.startswith("#"):
                continue
            l = l.split(" ")[:3]
            uni = int(l[1], 16)
            _by_uni[uni] = l[0]
            _by_glyph[l[0]] = uni
            _class_lookup[l[0]] = l[2]
    #except:
    #    pass

def uni_to_glyph(u):
    if not _by_uni:
        _populate_glyphs_unis()
    return _by_uni.get(u)
    
def glyph_to_uni(g):
    if g.lower() in [
        "gcommaaccent",
        "kcommaaccent",
        "lcommaaccent",
        "ncommaaccent",
        "rcommaaccent",
        ]:
        g = g.replace("commaaccent", "cedilla")
    elif g.lower() == "kgreenlandic":
        g = g.replace("greenlandic", "ra")
    if not _by_glyph:
        _populate_glyphs_unis()
    return _by_glyph.get(g)

def glyph_to_class(g):
    if not _class_lookup:
        _populate_glyphs_unis()
    return _class_lookup.get(g)


from coldtype.geometry.geometrical import Geometrical


class Atom(Geometrical):
    def __init__(self, x):
        self.x = x
    
    __hash__ = object.__hash__
    
    def __eq__(self, o):
        # TODO isclose?
        try:
            return self.x == o.x
        except AttributeError:
            return self.x == o

    def __repr__(self):
        return f"Atom({self.x})"

    def __getitem__(self, key):
        if key == 0:
            return self.x
        else:
            raise TypeError("Index must be 0")
    
    def __len__(self):
        return 1

    def __setitem__(self, key, value):
        if key == 0:
            self.x = value
        else:
            raise IndexError(
                "Invalid index for atom assignment, must be 0")
    
    def reverse(self):
        return self


from coldtype.img.datimage import DATImage
from coldtype.fx.skia import precompose
from coldtype.geometry import Rect
from coldtype.pens.datpen import DATPen, DATPens
import skia

class SkiaImage(DATImage):    
    def load_image(self, src):
        return skia.Image.MakeFromEncoded(skia.Data.MakeFromFileName(str(src)))
    
    def width(self):
        return self._img.width()
    
    def height(self):
        return self._img.height()

    def _resize(self, fx, fy):
        self._img = self._img.resize(
            int(self._img.width()*fx),
            int(self._img.height()*fy))
    
    def _precompose_fn(self):
        return precompose


from collections import defaultdict
import functools
import io
import os
import pathlib
import pickle
import sys
import tempfile
from types import SimpleNamespace

try:
    import numpy
except ImportError:
    pass

from fontTools.pens.basePen import BasePen
from fontTools.pens.pointPen import PointToSegmentPen
from fontTools.pens.recordingPen import RecordingPen
from fontTools.designspaceLib import DesignSpaceDocument
from fontTools.ttLib import TTFont
from fontTools.ufoLib import UFOReader
from fontTools.varLib.models import normalizeValue
from .baseFont import BaseFont
from .glyphDrawing import GlyphDrawing
from .ufoFont import Glyph, NotDefGlyph, UFOState, extractIncludedFeatureFiles
from ..compile.compilerPool import compileUFOToPath, compileDSToBytes
from ..compile.dsCompiler import getTTPaths
from ..misc.hbShape import HBShape
from ..misc.properties import cachedProperty
try:
    from ..mac.makePathFromOutline import makePathFromArrays
except:
    pass


class DesignSpaceSourceError(Exception):
    pass


class DSFont(BaseFont):

    def __init__(self, fontPath, fontNumber, dataProvider=None):
        super().__init__(fontPath, fontNumber)
        self.doc = None
        self._varGlyphs = {}
        self._normalizedLocation = {}
        self._sourceFontData = {}
        self._ufos = {}
        self._needsVFRebuild = True

    def resetCache(self):
        super().resetCache()
        self._varGlyphs = {}
        del self.defaultInfo
        del self.defaultVerticalAdvance
        del self.defaultVerticalOriginY

    def load(self, outputWriter):
        if self.doc is None:
            self.doc = DesignSpaceDocument.fromfile(self.fontPath)
            self.doc.findDefault()

        with tempfile.TemporaryDirectory(prefix="fontgoggles_temp") as ttFolder:
            sourcePathToTTPath = getTTPaths(self.doc, ttFolder)
            ufosToCompile = []
            ttPaths = []
            outputs = []
            coros = []
            self._sourceFiles = defaultdict(list)
            self._includedFeatureFiles = defaultdict(list)
            previousUFOs = self._ufos
            self._ufos = {}
            previousSourceData = self._sourceFontData
            self._sourceFontData = {}

            for source in self.doc.sources:
                sourceKey = (source.path, source.layerName)
                self._sourceFiles[pathlib.Path(source.path)].append(sourceKey)
                ufoState = previousUFOs.get(sourceKey)
                if ufoState is None:
                    reader = UFOReader(source.path, validate=False)
                    glyphSet = reader.getGlyphSet(layerName=source.layerName)
                    glyphSet.glyphClass = Glyph
                    if source.layerName is None:
                        includedFeatureFiles = extractIncludedFeatureFiles(source.path, reader)
                        getUnicodesAndAnchors = functools.partial(self._getUnicodesAndAnchors, source.path)
                    else:
                        includedFeatureFiles = []
                        # We're not compiling features nor do we need cmaps for these sparse layers,
                        # so we don't need need proper anchor or unicode data
                        def getUnicodesAndAnchors(): return ({}, {})
                    ufoState = UFOState(reader, glyphSet,
                                        getUnicodesAndAnchors=getUnicodesAndAnchors,
                                        includedFeatureFiles=includedFeatureFiles)
                for includedFeaFile in ufoState.includedFeatureFiles:
                    self._includedFeatureFiles[includedFeaFile].append(sourceKey)
                self._ufos[sourceKey] = ufoState

                if source.layerName is not None:
                    continue

                if source.path in ufosToCompile:
                    continue
                ttPath = sourcePathToTTPath[source.path]
                if source.path in previousSourceData:
                    with open(ttPath, "wb") as f:
                        f.write(previousSourceData[source.path])
                    self._sourceFontData[source.path] = previousSourceData[source.path]
                else:
                    ufosToCompile.append(source.path)
                    ttPaths.append(ttPath)
                    output = io.StringIO()
                    outputs.append(output)
                    coros.append(compileUFOToPath(source.path, ttPath, None))

            errors = []

            for sourcePath, exc, output in zip(ufosToCompile, errors, outputs):
                output = output.getvalue()
                if output or exc is not None:
                    outputWriter(f"compile output for {sourcePath}:\n")
                    if output:
                        outputWriter(output)
                    if exc is not None:
                        outputWriter(f"{exc!r}\n")

            if any(errors):
                raise DesignSpaceSourceError(
                    f"Could not build '{os.path.basename(self.fontPath)}': "
                    "some sources did not successfully compile"
                )
            for sourcePath, ttPath in zip(ufosToCompile, ttPaths):
                # Store compiled tt data so we can reuse it to rebuild ourselves
                # without recompiling the source.
                with open(ttPath, "rb") as f:
                    self._sourceFontData[sourcePath] = f.read()

            if not ufosToCompile and not self._needsVFRebuild:
                # self.ttFont and self.shaper are still up-to-date
                return

            vfFontData = compileDSToBytes(self.fontPath, ttFolder, outputWriter)

        f = io.BytesIO(vfFontData)
        self.ttFont = TTFont(f, lazy=True)
        # Nice cookie for us from the worker
        self.masterModel = pickle.loads(self.ttFont["MPcl"].data)
        assert len(self.masterModel.deltaWeights) == len(self.doc.sources)

        self.shaper = HBShape(vfFontData,
                            getHorizontalAdvance=self._getHorizontalAdvance,
                            getVerticalAdvance=self._getVerticalAdvance,
                            getVerticalOrigin=self._getVerticalOrigin,
                            ttFont=self.ttFont)
        self._needsVFRebuild = False

    def getExternalFiles(self):
        return sorted(self._sourceFiles) + sorted(self._includedFeatureFiles)

    def canReloadWithChange(self, externalFilePath):
        invalidateCaches = False
        if not externalFilePath:
            # Our .designspace file itself changed, let's reload
            self.doc = None
            self._needsVFRebuild = True
            invalidateCaches = True
        else:
            for sourcePath, sourceLayerName in self._includedFeatureFiles.get(externalFilePath, ()):
                assert sourceLayerName is None
                self._sourceFontData.pop(sourcePath, None)  # implies self._needsVFRebuild
                invalidateCaches = True
            for sourcePath, sourceLayerName in self._sourceFiles.get(externalFilePath, ()):
                sourceKey = sourcePath, sourceLayerName
                self._ufos[sourceKey] = self._ufos[sourceKey].newState()
                (needsFeaturesUpdate, needsGlyphUpdate,needsInfoUpdate, needsCmapUpdate, needsLibUpdate) = self._ufos[sourceKey].getUpdateInfo()
                if sourceLayerName is not None:
                    # We don't compile features for layer masters
                    needsFeaturesUpdate = False
                if needsFeaturesUpdate:
                    self._sourceFontData.pop(sourcePath, None)  # implies self._needsVFRebuild
                    invalidateCaches = True
                if needsGlyphUpdate or needsInfoUpdate:
                    invalidateCaches = True
                if needsCmapUpdate:
                    # TODO: This could be done more efficiently like how UFOFont
                    # does it, if the changed source is the default source.
                    self.doc = None
                    self._needsVFRebuild = True
                    invalidateCaches = True
        if invalidateCaches:
            self.resetCache()
        return True

    @cachedProperty
    def defaultInfo(self):
        info = SimpleNamespace()
        reader = self._ufos[(self.doc.default.path, self.doc.default.layerName)].reader
        reader.readInfo(info)
        return info

    @cachedProperty
    def unitsPerEm(self):
        return self.defaultInfo.unitsPerEm

    @cachedProperty
    def defaultVerticalAdvance(self):
        ascender = getattr(self.defaultInfo, "ascender", None)
        descender = getattr(self.defaultInfo, "descender", None)
        if ascender is None or descender is None:
            return self.defaultInfo.unitsPerEm
        else:
            return ascender + abs(descender)

    @cachedProperty
    def defaultVerticalOriginY(self):
        ascender = getattr(self.defaultInfo, "ascender", None)
        if ascender is None:
            return self.defaultInfo.unitsPerEm  # ???
        else:
            return ascender

    def varLocationChanged(self, varLocation):
        self._normalizedLocation = normalizeLocation(self.doc, varLocation or {})

    def _getVarGlyph(self, glyphName):
        varGlyph = self._varGlyphs.get(glyphName)
        if varGlyph is None:
            if glyphName not in self._ufos[(self.doc.default.path, self.doc.default.layerName)].glyphSet:
                varGlyph = NotDefGlyph(self.unitsPerEm)
            else:
                varGlyph = self._getVarGlyphRaw(glyphName)
            self._varGlyphs[glyphName] = varGlyph
        varGlyph.setVarLocation(self._normalizedLocation)
        return varGlyph

    def _getVarGlyphRaw(self, glyphName):
        tags = None
        contours = None
        components = None
        getSubGlyph = None
        masterPoints = []
        for source in self.doc.sources:
            glyphSet = self._ufos[(source.path, source.layerName)].glyphSet
            if glyphName not in glyphSet:
                masterPoints.append(None)
                continue
            glyph = glyphSet[glyphName]
            coll = PointCollector(glyphSet)
            try:
                glyph.draw(coll)
                if coll.points and coll.components:
                    # When the source mixes outlines and component we need
                    # to decompose to match fontmake/TT behavior
                    coll = PointCollector(glyphSet, decompose=True)
                    glyph.draw(coll)
            except Exception as e:
                print(f"Glyph '{glyphName}' could not be read from '{os.path.basename(source.path)}': {e!r}",
                      file=sys.stderr)
                masterPoints.append(None)
            else:
                hAdvance = glyph.width
                vAdvance = glyph.height
                if vAdvance is None or vAdvance == 0:  # XXX default vAdv == 0 -> bad UFO spec
                    vAdvance = self.defaultVerticalAdvance
                vOrgX = hAdvance / 2
                vOrgY = getattr(glyph, "lib", {}).get("public.verticalOrigin")
                if vOrgY is None:
                    vOrgY = self.defaultVerticalOriginY
                phantomPoints = [(hAdvance, 0), (vOrgX, vOrgY), (vOrgX, vOrgY - vAdvance)]
                if coll.components:
                    # Use the component offsets as points (the 2x2 matrix won't interpolate anyway)
                    points = [t[4:6] for bgn, t in coll.components]
                else:
                    points = coll.points
                masterPoints.append(points + phantomPoints)
                if source is self.doc.default:
                    tags = coll.tags
                    contours = coll.contours
                    components = coll.components
                    getSubGlyph = self._getVarGlyph

        if tags is None:
            print(f"Default master glyph '{glyphName}' could not be read", file=sys.stderr)
            varGlyph = NotDefGlyph(self.unitsPerEm)
        else:
            varGlyph = VarGlyph(glyphName, self.masterModel, masterPoints, contours, tags,
                                components, getSubGlyph)
        return varGlyph

    def _getHorizontalAdvance(self, glyphName):
        varGlyph = self._getVarGlyph(glyphName)
        return varGlyph.width

    def _getVerticalAdvance(self, glyphName):
        varGlyph = self._getVarGlyph(glyphName)
        return -abs(varGlyph.height)

    def _getVerticalOrigin(self, glyphName):
        varGlyph = self._getVarGlyph(glyphName)
        vOrgX, vOrgY = varGlyph.verticalOrigin
        return True, vOrgX, vOrgY

    def _getGlyphDrawing(self, glyphName, colorLayers):
        try:
            varGlyph = self._getVarGlyph(glyphName)
            return GlyphDrawing([(varGlyph.getOutline(cocoa=self.cocoa), None)])
        except Exception as e:
            print(f"Can't get outline for '{glyphName}': {e!r}", file=sys.stderr)
            return GlyphDrawing([])

    def _getUnicodesAndAnchors(self, sourcePath):
        f = io.BytesIO(self._sourceFontData[sourcePath])
        ttFont = TTFont(f, lazy=True)
        unicodes = defaultdict(list)
        for code, gn in ttFont.getBestCmap().items():
            unicodes[gn].append(code)
        anchors = pickle.loads(ttFont["FGAx"].data)
        return unicodes, anchors


# From FreeType:
FT_CURVE_TAG_ON = 1
FT_CURVE_TAG_CONIC = 0
FT_CURVE_TAG_CUBIC = 2

segmentTypes = {FT_CURVE_TAG_ON: "line", FT_CURVE_TAG_CONIC: "qcurve", FT_CURVE_TAG_CUBIC: "curve"}
coordinateType = float


def interpolateFromDeltas(model, varLocation, deltas):
    # This is a numpy-specific reimplementation of model.interpolateFromDeltas()
    # that avoids allocation of in-between results.
    # However, so far this is achieving the speedup I had hoped...
    # Perhaps there'll be a better improvement if there are many deltas
    # and/or if the outlines are more complex.
    deltas = deltas
    temp = numpy.zeros(deltas[0].shape, coordinateType)
    v = numpy.zeros(deltas[0].shape, coordinateType)
    scalars = model.getScalars(varLocation)
    for delta, scalar in zip(deltas, scalars):
        if not scalar:
            continue
        if scalar == 1.0:
            contribution = delta
        else:
            numpy.multiply(delta, scalar, temp)
            contribution = temp
        numpy.add(v, contribution, v)
    return v


NUMPY_IN_PLACE = True  # dubious improvement


class VarGlyph:

    def __init__(self, glyphName, masterModel, masterPoints, contours, tags, components, getSubGlyph):
        self.model, masterPoints = masterModel.getSubModel(masterPoints)
        masterPoints = [numpy.array(pts, coordinateType) for pts in masterPoints]
        try:
            self.deltas = self.model.getDeltas(masterPoints)
        except ValueError:
            # outlines are not compatible, fall back to the default master
            print(f"Glyph '{glyphName}' is not interpolatable", file=sys.stderr)
            self.deltas = [masterPoints[self.model.reverseMapping[0]]]
        if components:
            self._contours = None
            self._tags = None
        else:
            self._contours = numpy.array(contours, numpy.short)
            self._tags = numpy.array(tags, numpy.byte)
        self.components = components
        self._getSubGlyph = getSubGlyph
        self.varLocation = {}
        self._points = None

    def setVarLocation(self, varLocation):
        if varLocation is None:
            varLocation = {}
        if self.varLocation == varLocation:
            return
        self._points = None
        self.varLocation = varLocation

    @property
    def contours(self):
        if self._contours is None:
            firstPoint = 0
            allContours = []
            for glyphName, transformation in self.components:
                subGlyph = self._getSubGlyph(glyphName)
                if isinstance(subGlyph, NotDefGlyph):
                    continue
                allContours.append(subGlyph.contours + firstPoint)
                firstPoint = subGlyph.contours[-1] + firstPoint + 1
            self._contours = numpy.concatenate(allContours)
        return self._contours

    @property
    def tags(self):
        if self._tags is None:
            allTags = []
            for glyphName, transformation in self.components:
                subGlyph = self._getSubGlyph(glyphName)
                if isinstance(subGlyph, NotDefGlyph):
                    continue
                allTags.append(subGlyph.tags)
            self._tags = numpy.concatenate(allTags)
        return self._tags

    def getPoints(self):
        if self._points is None:
            if NUMPY_IN_PLACE:
                self._points = interpolateFromDeltas(self.model, self.varLocation, self.deltas)
            else:
                self._points = self.model.interpolateFromDeltas(self.varLocation, self.deltas)

            if self.components:
                allPoints = []
                for (glyphName, transformation), offset in zip(self.components, self._points):
                    twoByTwo = transformation[:4]
                    subGlyph = self._getSubGlyph(glyphName)
                    if isinstance(subGlyph, NotDefGlyph):
                        print(f"Composite base glyph '{glyphName}' not found", file=sys.stderr)
                        continue
                    subPoints = subGlyph.getPoints()[:-3]  # strip phantom points
                    if twoByTwo != (1, 0, 0, 1):  # identity
                        m = [twoByTwo[:2], twoByTwo[2:]]
                        subPoints = subPoints @ m  # matrix multiply
                    allPoints.append(subPoints + offset)  # skip phantom points
                allPoints.append(self._points[-3:])  # add phantom points
                self._points = numpy.concatenate(allPoints)

        return self._points

    @property
    def width(self):
        return self.getPoints()[-3][0]

    @property
    def height(self):
        points = self.getPoints()
        return points[-1][1] - points[-2][1]

    @property
    def verticalOrigin(self):
        return self.getPoints()[-2]

    def getOutline(self, cocoa=True):
        if cocoa:
            return makePathFromArrays(self.getPoints(), self.tags, self.contours)
        else:
            #print(self.tags)
            rp = RecordingPen()
            self.draw(rp)
            return rp

    def draw(self, pen):
        ppen = PointToSegmentPen(pen)
        startIndex = 0
        points = self.getPoints()
        for endIndex in self.contours:
            lastTag = self.tags[endIndex]
            endIndex += 1
            contourTags = self.tags[startIndex:endIndex]
            contourPoints = points[startIndex:endIndex]
            ppen.beginPath()
            for tag, (x, y) in zip(contourTags, contourPoints):
                if tag == FT_CURVE_TAG_ON:
                    segmentType = segmentTypes[lastTag]
                else:
                    segmentType = None
                ppen.addPoint((x, y), segmentType=segmentType)
                lastTag = tag
            ppen.endPath()
            startIndex = endIndex


class PointCollector(BasePen):

    def __init__(self, glyphSet, decompose=False):
        super().__init__(glyphSet)
        self.decompose = decompose
        self.points = []
        self.tags = []
        self.contours = []
        self.components = []
        self.contourStartPointIndex = None

    def moveTo(self, pt):
        self.contourStartPointIndex = len(self.points)
        self.points.append(pt)
        self.tags.append(FT_CURVE_TAG_ON)

    def lineTo(self, pt):
        self.points.append(pt)
        self.tags.append(FT_CURVE_TAG_ON)

    def curveTo(self, *pts):
        self.tags.extend([FT_CURVE_TAG_CUBIC] * (len(pts) - 1))
        self.tags.append(FT_CURVE_TAG_ON)
        self.points.extend(pts)

    def qCurveTo(self, *pts):
        self.tags.extend([FT_CURVE_TAG_CONIC] * (len(pts) - 1))
        if pts[-1] is None:
            self.contourStartPointIndex = len(self.points)
            pts = pts[:-1]
        else:
            self.tags.append(FT_CURVE_TAG_ON)
        self.points.extend(pts)

    def closePath(self):
        assert self.contourStartPointIndex is not None
        currentPointIndex = len(self.points) - 1
        if (self.contourStartPointIndex != currentPointIndex and
                self.points[self.contourStartPointIndex] == self.points[currentPointIndex] and
                self.tags[self.contourStartPointIndex] == self.tags[currentPointIndex]):
            self.points.pop()
            self.tags.pop()
        self.contours.append(len(self.points) - 1)
        self.contourStartPointIndex = None

    endPath = closePath

    def addComponent(self, glyphName, transformation):
        if self.decompose:
            super().addComponent(glyphName, transformation)
        else:
            self.components.append((glyphName, transformation))


def normalizeLocation(doc, location):
    # Adapted from DesignSpaceDocument.normalizeLocation(), which takes axis
    # names, yet we need to work with tags here.
    # Also, the original takes design space coordinates, whereas I have user
    # space coordinates here.
    new = {}
    for axis in doc.axes:
        if axis.tag not in location:
            # skipping this dimension it seems
            continue
        value = axis.map_forward(location[axis.tag])
        triple = [
            axis.map_forward(v) for v in (axis.minimum, axis.default, axis.maximum)
        ]
        new[axis.tag] = normalizeValue(value, triple)
    return new



import math
import os.path

try:
    import numpy as np
except ImportError:
    np = None

try:
    import soundfile as sf
except:
    sf = None

from coldtype.pens.datpen import DATPen

class Wavfile():
    def __init__(self, path, fps=30):
        self.sf, self.sf_fs = sf.read(str(path))
        self.fps = fps
        self.hz = self.sf_fs
        self.samples_per_frame = self.hz / fps
        self.path = path
        self.peaks = self.sf
        self.framelength = int(round(len(self.peaks) / self.samples_per_frame))

        max_frame_amp = 0
        for i in range(0, self.framelength):
            amp = self.amp(i)
            if amp > max_frame_amp:
                max_frame_amp = amp
        self.max_frame_amp = max_frame_amp

    def calc_peaks(self):
        snd = self.sf / (2.**15)
        s1 = snd[:, 0]
        return s1

    def samples_for_frame(self, i):
        start_sample = math.floor(i * self.samples_per_frame)
        end_sample = math.floor(start_sample + self.samples_per_frame)
        return self.peaks[start_sample:end_sample]

    def amp(self, i):
        return np.average(np.fabs(self.samples_for_frame(i)))
    
    def frame_waveform(self, fi, r, inc=1, pen=None):
        wave = pen or DATPen()
        samples = self.samples_for_frame(fi)[::inc]
        ww = r.w/len(samples)
        wh = r.h
        for idx, w in enumerate(samples):
            if idx == 0:
                wave.moveTo((r.x, w[0]*wh))
            else:
                wave.lineTo((idx*ww, w[0]*wh))
        wave.endPath()
        wave.f(None).s(1).sw(2)
        return wave



from coldtype.time.easing import ease
from copy import copy
import math


class Timing():
    def __init__(self, t, loop_t, loop, easefn):
        self.t = t
        self.loop_t = loop_t
        self.loop = loop
        self.loop_phase = int(loop%2 != 0)
        self.e, self.s = self.ease(easefn)
    
    def ease(self, easefn):
        easer = easefn
        if not isinstance(easer, str) and not hasattr(easer, "value"):
            try:
                iter(easefn) # is-iterable
                if len(easefn) > self.loop:
                    easer = easefn[self.loop]
                elif len(easefn) == 2:
                    easer = easefn[self.loop % 2]
                elif len(easefn) == 1:
                    easer = easefn[0]
                else:
                    easer = easefn[0]
            except TypeError:
                print("failed")
                pass
        v, tangent = ease(easer, self.loop_t)
        return min(1, max(0, v)), tangent


class Timeable():
    """
    Abstract base class for anything with a concept of `start` and `end`/`duration`

    Implements additional methods to make it easier to work with time-based concepts
    """
    def __init__(self, start, end, index=0, name=None, data={}, timeline=None):
        self.start = start
        self.end = end
        self.index = index
        self.name = name
        self.feedback = 0
        self.data = data
        self.timeline = timeline
    
    @property
    def duration(self):
        return self.end - self.start
    
    def __repr__(self):
        return f"Timeable({self.start}, {self.end} ({self.duration}))"
    
    def delay(self, frames_delayed, feedback) -> 'Timeable':
        t = copy(self)
        t.start = t.start + frames_delayed
        t.end = t.end + frames_delayed
        t.feedback = feedback
        t.data = {}
        return t
    
    def retime(self, start=0, end=0, duration=-1):
        self.start = self.start + start
        self.end = self.end + end
        if duration > -1:
            self.end = self.start + duration
        return self
    
    def now(self, i):
        return self.start <= i < self.end

    def _normalize_fi(self, fi):
        if self.timeline:
            if self.end > self.timeline.duration and fi < self.start:
                return fi + self.timeline.duration
        return fi
    
    def e(self, fi, easefn="eeio", loops=1, rng=(0, 1), cyclic=True, to1=False, out1=False):
        if not isinstance(easefn, str):
            loops = easefn
            easefn = "eeio"
        fi = self._normalize_fi(fi)
        t = self.progress(fi, loops=loops, easefn=easefn, cyclic=cyclic, to1=to1, out1=out1)
        e = t.e
        ra, rb = rng
        if ra > rb:
            e = 1 - e
            rb, ra = ra, rb
        return ra + e*(rb - ra)

    def io(self, fi, length, ei="eei", eo="eei", negative=False):
        """
        Somewhat like ``progress()``, but can be used to fade in/out (hence the name (i)n/(o)ut)

        * ``length`` refers to the lenght of the ease, in frames
        * ``ei=`` takes the ease-in mnemonic
        * ``eo=`` takes the ease-out mnemonic
        """
        try:
            length_i, length_o = length
        except:
            length_i = length
            length_o = length
        
        fi = self._normalize_fi(fi)

        if fi < self.start:
            return 1
        if fi > self.end:
            return 0
        to_end = self.end - fi
        to_start = fi - self.start
        easefn = None
        in_end = False
        if to_end < length_o and eo:
            in_end = True
            v = 1-to_end/length_o
            easefn = eo
        elif to_start <= length_i and ei:
            v = 1-to_start/length_i
            easefn = ei
        else:
            v = 0
        if v == 0 or not easefn:
            return 0
        else:
            a, _ = ease(easefn, v)
            if negative and in_end:
                return -a
            else:
                return a
    
    def io2(self, fi, length, easefn="eeio", negative=False):
        try:
            length_i, length_o = length
        except:
            length_i = length
            length_o = length
        
        if isinstance(length_i, float):
            length_i = int(self.duration*(length_i/2))
        if isinstance(length_o, float):
            length_o = int(self.duration*(length_o/2))
        
        if fi < self.start or fi > self.end:
            return 0
        
        try:
            ei, eo = easefn
        except ValueError:
            ei, eo = easefn, easefn

        to_end = self.end - fi
        to_start = fi - self.start
        easefn = None
        in_end = False

        if to_end < length_o and eo:
            in_end = True
            v = to_end/length_o
            easefn = eo
        elif to_start <= length_i and ei:
            v = to_start/length_i
            easefn = ei
        else:
            v = 1

        if v == 1 or not easefn:
            return 1
        else:
            a, _ = ease(easefn, v)
            return a
            if negative and in_end:
                return -a
            else:
                return a

    def _loop(self, t, times=1, cyclic=True, negative=False):
        lt = t*times*2
        ltf = math.floor(lt)
        ltc = math.ceil(lt)
        if False:
            if ltc % 2 != 0: # looping back
                lt = 1 - (ltc - lt)
            else: # looping forward
                lt = ltc - lt
        lt = lt - ltf
        if cyclic and ltf%2 == 1:
            if negative:
                lt = -lt
            else:
                lt = 1 - lt
        return lt, ltf
    
    def progress(self, i, loops=0, cyclic=True, negative=False, easefn="linear", to1=False, out1=True) -> Timing:
        """
        Given an easing function (``easefn=``), calculate the amount of progress as a Timing object

        ``easefn=`` takes a mnemonic as enumerated in :func:`coldtype.time.easing.ease`
        """
        if i < self.start:
            return Timing(0, 0, 0, easefn)
        if i > self.end:
            if out1:
                return Timing(1, 1, 0, easefn)
            else:
                return Timing(0, 0, 0, easefn)
        d = self.duration
        if to1:
            d = d - 1
        t = (i-self.start) / d
        if loops == 0:
            return Timing(t, t, 0, easefn)
        else:
            loop_t, loop_index = self._loop(t, times=loops, cyclic=cyclic, negative=negative)
            return Timing(t, loop_t, loop_index, easefn)
    
    def halfover(self, i):
        e = self.progress(i, to1=1).e
        return e >= 0.5
    
    #prg = progress


class TimeableView(Timeable):
    def __init__(self, timeable, value, svalue, count, index, position, start, end):
        self.timeable = timeable
        self.value = value
        self.svalue = svalue
        self.count = count
        self.index = index
        self.position = position
        self.start = start
        self.end = end
        super().__init__(start, end)
    
    def ease(self, eo="eei", ei="eei"):
        return ease(eo, self.value)[0]
    
    def __repr__(self):
        return f"<TimeableView:{self.timeable}/>"


class TimeableSet():
    def __init__(self, timeables, name=None, start=-1, end=-1, data={}):
        self.timeables = timeables
        self.name = name
        self._start = start
        self._end = end
        self.data = data
    
    def flat_timeables(self):
        ts = []
        for t in self.timeables:
            if isinstance(t, TimeableSet):
                ts.extend(t.flat_timeables())
            else:
                ts.append(t)
        return ts

    def constrain(self, start, end):
        self._start = start
        self._end = end
    
    @property
    def start(self):
        if self._start > -1:
            return self._start
        _start = -1
        for t in self.timeables:
            ts = t.start
            if _start == -1:
                _start = ts
            elif ts < _start:
                _start = ts
        return _start

    @property
    def end(self):
        if self._end > -1:
            return self._end
        _end = -1
        for t in self.timeables:
            te = t.end
            if _end == -1:
                _end = te
            elif te > _end:
                _end = te
        return _end
    
    def __getitem__(self, index):
        return self.timeables[index]
    
    def current(self, frame):
        for idx, t in enumerate(self.flat_timeables()):
            t:Timeable
            if t.start <= frame and frame < t.end:
                return t
    
    def fv(self, frame, filter_fn=None, reverb=[0,5], duration=-1, accumulate=0):
        pre, post = reverb
        count = 0
        timeables_on = []

        ts_duration = self.end - self.start

        for idx, t in enumerate(self.flat_timeables()):
            if filter_fn and not filter_fn(t):
                continue
            t_start = t.start
            t_end = t.end
            if duration > -1:
                t_end = t_start + duration
            pre_start = t_start - pre
            post_end = t_end + post

            t_index = count
            if frame >= pre_start: # correct?
                count += 1
            
            value = 0
            pos = 0
            fi = frame

            if frame >= t_start and frame <= t_end: # truly on
                pos = 0
                value = 1
            else:
                if post_end > ts_duration and frame < pre_start:
                    fi = frame + ts_duration
                elif pre_start < 0 and frame > post_end:
                    fi = frame - ts_duration
                if fi < t_start and fi >= pre_start:
                    pos = 1
                    value = (fi - pre_start) / pre
                elif fi > t_end and fi < post_end:
                    pos = -1
                    value = (post_end - fi) / post
            
            if value > 0:
                timeables_on.append(TimeableView(t, value, -1, count, idx, pos, pre_start, post_end))
            else:
                pass
        
        if accumulate:
            return timeables_on
        else:
            if len(timeables_on) == 0:
                return TimeableView(None, 0, 0, count, -1, 0, 0, 0)
            else:
                return max(timeables_on, key=lambda tv: tv.value)

    def __repr__(self):
        return "<TimeableSet ({:s}){:04d}>".format(self.name if self.name else "?", len(self.timeables))


import math, sys, os, re
from pathlib import Path

from coldtype.text import *
from coldtype.text.reader import Font
from coldtype.pens.datpen import DATPen, DATPens, DATPenSet, DP, DPS
from coldtype.pens.dattext import DATText
from coldtype.geometry import *
from coldtype.color import *
from coldtype.renderable import *
from coldtype.helpers import loopidx, sibling, raw_ufo, ßhide, ßshow, cycle_idx, random_series, show_points, glyph_to_uni, uni_to_glyph, glyph_to_class, DefconFont
from coldtype.time import *
from coldtype.img.blendmode import BlendMode
#from coldtype.renderer.state import RendererState, Keylayer

name = "coldtype"
__version__ = "0.5.2"

__sibling__ = lambda x: x # will be redefined contextually

def noop():
    return None


import tempfile, traceback, threading, multiprocessing
import argparse, importlib, inspect, json, math
import sys, os, re, signal, tracemalloc, shutil
import platform, pickle, string, datetime

try:
    import numpy as np
except ImportError:
    np = None

import time as ptime
from pathlib import Path
from typing import Tuple
from pprint import pprint
from runpy import run_path
from subprocess import Popen, PIPE, STDOUT
from random import shuffle, Random
from more_itertools import distribute
from docutils.core import publish_doctree
from functools import partial

from http.server import HTTPServer, SimpleHTTPRequestHandler

import coldtype
from coldtype.helpers import *
from coldtype.geometry import Rect, Point, Edge
from coldtype.text.reader import Font

from coldtype.renderer.state import RendererState, Keylayer, Overlay
from coldtype.renderable import renderable, Action, animation
from coldtype.pens.datpen import DATPen, DATPens

from coldtype.pens.svgpen import SVGPen
from coldtype.pens.jsonpen import JSONPen

from coldtype.renderer.keyboard import KeyboardShortcut, SHORTCUTS, REPEATABLE_SHORTCUTS, symbol_to_glfw

try:
    from coldtype.renderer.watchdog import AsyncWatchdog
except ImportError:
    AsyncWatchdog = None

try:
    import skia
    from coldtype.pens.skiapen import SkiaPen
    import coldtype.fx.skia as skfx
except ImportError:
    skia = None
    SkiaPen = None

try:
    import drawBot as db
except ImportError:
    db = None

try:
    import glfw
except ImportError:
    glfw = None

# source: https://github.com/PixarAnimationStudios/USD/issues/1372

def monkeypatch_ctypes():
    import os
    import ctypes.util
    uname = os.uname()
    if uname.sysname == "Darwin" and uname.release >= "20.":
        real_find_library = ctypes.util.find_library
        def find_library(name):
            if name in {"OpenGL", "GLUT"}:  # add more names here if necessary
                return f"/System/Library/Frameworks/{name}.framework/{name}"
            return real_find_library(name)
        ctypes.util.find_library = find_library
    return

try:
    from OpenGL import GL
except ImportError:
    monkeypatch_ctypes()
    try:
        from OpenGL import GL
    except:
        print("pip install PyOpenGL")
        GL = None

from coldtype.renderer.utils import *

_random = Random()

from time import sleep

try:
    import rtmidi
except ImportError:
    pass

try:
    import pyaudio, wave, soundfile
except ImportError:
    pyaudio = None
    wave = None

try:
    import psutil
    process = psutil.Process(os.getpid())
except ImportError:
    process = None

DARWIN = platform.system() == "Darwin"

last_line = ''
new_line_event = threading.Event()

def monitor_stdin():
    # https://stackoverflow.com/questions/27174736/how-to-read-most-recent-line-from-stdin-in-python
    global last_line
    global new_line_event

    def keep_last_line():
        global last_line, new_line_event
        for line in sys.stdin:
            last_line = line
            new_line_event.set()

    keep_last_line_thread = threading.Thread(target=keep_last_line)
    keep_last_line_thread.daemon = True
    keep_last_line_thread.start()

def show_tail(p):
    for line in p.stdout:
        print(line.decode("utf-8").split(">>")[-1].strip().strip("\n"))

WEBSOCKET_PORT = None

class WebViewerHandler(SimpleHTTPRequestHandler):
    def _set_headers(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def do_GET(self):
        self._set_headers()
        self.wfile.write(
            (Path(__file__).parent.parent / "webserver/webviewer.html")
                .read_text()
                .replace("localhost:8007", f"localhost:{WEBSOCKET_PORT}")
                .encode("utf8"))

    def do_HEAD(self):
        self._set_headers()
    
    def log_message(self, format, *args):
        pass

class DataSourceThread(threading.Thread):
    def __init__(self, filepath):
        self.filepath = filepath
        self.program = run_path(self.filepath)
        self.should_run = True
        threading.Thread.__init__(self)
    
    def run(self):
        print("Running data thread for:", self.filepath.relative_to(Path.cwd()))
        while self.should_run:
            self.program["run"]()
            self.interval = self.program.get("INTERVAL", 0.5)
            ptime.sleep(self.interval)

class Renderer():
    def Argparser(name="coldtype", file=True, defaults={}, nargs=[]):
        parser = argparse.ArgumentParser(prog=name, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        
        if file:
            parser.add_argument("file", type=str, nargs="?", help="The source file for a coldtype render")
        for narg in nargs:
            parser.add_argument(narg[0], nargs="?", default=narg[1])
        
        pargs = dict(
            version=parser.add_argument("-v", "--version", action="store_true", default=False, help="Display version"),

            no_watch=parser.add_argument("-nw", "--no-watch", action="store_true", default=False, help="Preventing watching for changes to source files"),

            no_viewer=parser.add_argument("-nv", "--no-viewer", action="store_true", default=False, help="Prevent showing skia/glfw viewer"),

            websocket=parser.add_argument("-ws", "--websocket", action="store_true", default=False, help="Should the server run a web socket?"),

            webviewer=parser.add_argument("-wv", "--webviewer", action="store_true", default=False, help="Do you want to live-preview in a webviewer instead of with a glfw-skia window?"),

            webviewer_port=parser.add_argument("-wvp", "--webviewer-port", type=int, default=8008, help="What port should the webviewer run on? (provided -wv is passed)"),

            port=parser.add_argument("-p", "--port", type=int, default=8007, help="What port should the websocket run on (provided -ws is passed)"),
        
            websocket_external=parser.add_argument("-wse", "--websocket-external", type=str, default=None, help="What address should a websocket connection be made to, for control of external coldtype program?"),

            no_midi=parser.add_argument("-nm", "--no-midi", action="store_true", default=False, help="Midi is on by default, do you want to turn it off?"),
            
            save_renders=parser.add_argument("-sv", "--save-renders", action="store_true", default=False, help="Should the renderer create image artifacts?"),

            capture_previews=parser.add_argument("-cp", "--capture-previews", action="store_true", default=False, help="Should the viewer self-capture via connected cv2 camera?"),

            capture_previews_delay=parser.add_argument("-cpd", "--capture-previews-delay", type=float, default=0.35, help="How long should the rendering loop hang between displaying and capturing?"),
            
            rasterizer=parser.add_argument("-r", "--rasterizer", type=str, default=None, choices=["drawbot", "cairo", "svg", "skia", "pickle"], help="Which rasterization engine should coldtype use to create artifacts?"),

            cpu_render=parser.add_argument("-cpu", "--cpu-render", action="store_true", default=False, help="Should final rasters be performed without a GPU context?"),
            
            scale=parser.add_argument("-s", "--scale", type=float, default=1.0, help="When save-renders is engaged, what scale should images be rasterized at? (Useful for up-rezing)"),

            preview_scale=parser.add_argument("-ps", "--preview-scale", type=float, default=1.0, help="What size should previews be shown at?"),

            #preview_audio=parser.add_argument("-pa", "--preview-audio", action="store_true", default=False, help="Should the renderer attempt to playback audio?"),
            
            all=parser.add_argument("-a", "--all", action="store_true", default=False, help="If rendering an animation, pass the -a flag to render all frames sequentially"),

            build=parser.add_argument("-b", "--build", action="store_true", default=False, help="Should the build function be run and the renderer quit immediately?"),

            release=parser.add_argument("-rls", "--release", action="store_true", default=False, help="Should the release function be run and the renderer quit immediately?"),

            multiplex=parser.add_argument("-mp", "--multiplex", action="store_true", default=False, help="Render in multiple processes"),

            memory=parser.add_argument("-mm", "--memory", action="store_true", default=False, help="Show statistics about memory usage?"),

            midi_info=parser.add_argument("-mi", "--midi-info", action="store_true", default=False, help="Show available MIDI devices"),

            show_time=parser.add_argument("-st", "--show-time", action="store_true", default=False, help="Show time for each render"),

            is_subprocess=parser.add_argument("-isp", "--is-subprocess", action="store_true", default=False, help=argparse.SUPPRESS),

            thread_count=parser.add_argument("-tc", "--thread-count", type=int, default=defaults.get("thread_count", 8), help="How many threads when multiplexing?"),

            no_sound=parser.add_argument("-ns", "--no-sound", action="store_true", default=False, help="Don’t make sound"),

            window_opacity=parser.add_argument("-wo", "--window-opacity", type=float, help="opacity of the window, from >0 to 1 (defaults to 1 unless overridden in .coldtype.py file)"),

            monitor_name=parser.add_argument("-mn", "--monitor-name", type=str, help="the name of the monitor to open the window in; pass 'list' to list all monitor names"),

            window_pin=parser.add_argument("-wp", "--window-pin", type=str, help="where to pin the window, if you want to pin the window"),

            window_pin_inset=parser.add_argument("-wpi", "--window-pin-inset", type=str, help="how much to 'inset' the pin of the window"),

            window_content_scale=parser.add_argument("-wcs", "--window-content-scale", type=float, default=None, help="do you want to override the reported content-scale?"),

            window_float=parser.add_argument("-wf", "--window-float", action="store_true", default=False, help="should the window float on top of everything?"),

            window_transparent=parser.add_argument("-wt", "--window-transparent", action="store_true", default=False, help="should the window background be transparent?"),

            window_background=parser.add_argument("-wb", "--window-background", action="store_true", default=False, help="should the window not take focus when its first opened?"),

            window_passthrough=parser.add_argument("-wpass", "--window-passthrough", action="store_true", default=False, help="should the window ignore all mouse interaction?"),
            
            format=parser.add_argument("-fmt", "--format", type=str, default=None, help="What image format should be saved to disk?"),

            indices=parser.add_argument("-i", "--indices", type=str, default=None),

            output_folder=parser.add_argument("-of", "--output-folder", type=str, default=None, help="If you don’t want to render to the default output location, specify that here."),

            monitor_lines=parser.add_argument("-ml", "--monitor-lines", action="store_true", default=False, help=argparse.SUPPRESS),

            filter_functions=parser.add_argument("-ff", "--filter-functions", type=str, default=None, help="Names of functions to render"),

            sidecar=parser.add_argument("-sc", "--sidecar", type=str, default=None, help="A file to run alongside your coldtype source file (like a file that processes data or keystrokes), that will run in a managed thread"),

            tails=parser.add_argument("-tl", "--tails", type=str, default=None, help="File to tail, comma-separated (no whitespace); results will print output to the normal process output"),

            hide_keybuffer=parser.add_argument("-hkb", "--hide-keybuffer", action="store_true", default=False, help="Should the keybuffer be shown?"),

            show_exit_code=parser.add_argument("-sec", "--show-exit-code", action="store_true", default=False, help=argparse.SUPPRESS),

            show_render_count=parser.add_argument("-src", "--show-render-count", action="store_true", default=False, help=argparse.SUPPRESS),

            frame_offsets=parser.add_argument("-fo", "--frame-offsets", type=str, default=None, help=argparse.SUPPRESS),

            disable_syntax_mods=parser.add_argument("-dsm", "--disable-syntax-mods", action="store_true", default=False, help="Coldtype has some optional syntax modifications that require copying the source to a new tempfile before running — would you like to skip this to preserve __file__ in your sources?")),
        return pargs, parser
    
    def read_configs(self):
        proj = Path(".coldtype.py")
        user = Path("~/.coldtype.py").expanduser()
        self.midi_mapping = {}
        self.hotkey_mapping = {}
        self.py_config = {}
        self.many_increment = 10

        for p in [user, proj]:
            if p.exists():
                try:
                    self.py_config = {
                        **self.py_config,
                        **run_path(str(p), init_globals={
                            "__MIDI__": self.midi_mapping,
                            "__HOTKEYS__": self.hotkey_mapping,
                        })
                    }
                    self.midi_mapping = self.py_config.get("MIDI", self.midi_mapping)
                    self.hotkey_mapping = self.py_config.get("HOTKEYS", self.hotkey_mapping)
                    self.many_increment = self.py_config.get("MANY_INCREMENT", self.many_increment)
                except Exception as e:
                    print("Failed to load config", p)
                    print("Exception:", e)

    def __init__(self, parser, no_socket_ok=False):
        sys.path.insert(0, os.getcwd())

        self.read_configs()

        self.parser = parser
        self.args = parser.parse_args()
        if self.args.is_subprocess or self.args.all or self.args.release or self.args.build:
            self.args.no_watch = True
        
        if not self.args.webviewer and not skia and not glfw:
            print("No viewing renderer installed — rendering all...")
            sleep(1)
            #self.args.no_watch = True
            #self.args.all = True
        
        self.disable_syntax_mods = self.args.disable_syntax_mods
        self.filepath = None
        self.state = RendererState(self)
        self.state.capturing_previews = self.args.capture_previews

        self.observers = []
        self.watchees = []
        self.sidecar_threads = []
        self.tails = []
        self.watchee_mods = {}
        
        self.rasterizer_warning = None
        self.primary_monitor = None
        
        if not self.reset_filepath(self.args.file if hasattr(self.args, "file") else None):
            self.dead = True
            return
        else:
            self.dead = False
        
        self.state.preview_scale = self.args.preview_scale
        
        monitor_stdin()
        
        if self.args.version:
            print(">>>", coldtype.__version__)
            self.dead = True
            return

        self.program = None
        self.exit_code = 0
        self.all_shortcuts = self.shortcuts()

        self.line_number = -1
        self.last_renders = []
        self.last_render_cleared = False
        self.last_previews = []

        # for multiplex mode
        self.running_renderers = []
        self.completed_renderers = []

        self.action_waiting = None
        self.requests_waiting = []
        self.waiting_to_render = []
        self.previews_waiting_to_paint = []
        self.preloaded_frames = []
        self.playing_preloaded_frame = -1
        self.glfw_last_time = -1
        self.last_animation = None
        self.last_rect = None
        self.playing = 0
        self.hotkeys = None
        self.hotkey_waiting = None
        self.context = None
        self.surface = None
        self.transparent = False

        if self.args.filter_functions:
            self.function_filters = [f.strip() for f in self.args.filter_functions.split(",")]
        else:
            self.function_filters = []
        self.multiplexing = self.args.multiplex
        self._should_reload = False

        self.recurring_actions = {}
        self.viewer_solos = []
    
    def reset_filepath(self, filepath):
        for k, cv2cap in self.state.cv2caps.items():
            cv2cap.release()

        self.line_number = -1
        self.stop_watching_file_changes()
        self.state.input_history.clear()
        self.state._frame_offsets = {}
        self.state._initial_frame_offsets = {}
        self.state.cv2caps = {}

        if filepath:
            self.filepath = Path(filepath).expanduser().resolve()
            if self.filepath.name == "demo":
                self.filepath = Path(__file__).parent.parent / "demo/demo.py"
            elif self.filepath.suffix == ".py":
                if not self.filepath.exists(): # write some default code there after a prompt
                    print(">>> That python file does not exist...")
                    create = input(">>> Do you want to create it and add some coldtype boilerplate? (y/n): ")
                    if create.lower() == "y":
                        self.filepath.parent.mkdir(exist_ok=True, parents=True)
                        self.filepath.write_text("from coldtype import *\n\n@renderable()\ndef stub(r):\n    return (DATPen()\n        .oval(r.inset(50))\n        .f(0.8))\n")
                        editor_cmd = self.py_config.get("EDITOR_COMMAND")
                        if editor_cmd:
                            os.system(editor_cmd + " " + str(self.filepath.relative_to(Path.cwd())))
            elif self.filepath.suffix == ".rst":
                if not self.filepath.exists():
                    print(">>> That rst file does not exist")
                    return False
            else:
                print(">>> Coldtype can only read .py and .rst files")
                return False
            self.codepath = None
            self._codepath_offset = 0
            self.watchees = [[Watchable.Source, self.filepath, None]]
            if not self.args.is_subprocess:

                self.watch_file_changes()
                for line in self.filepath.read_text().splitlines():
                    if line.startswith("#coldtype"):
                        print("> Overriding command-line args with:\n    >", line)
                        self.args = self.parser.parse_args(line.replace("#coldtype", "").strip().split(" "))
        else:
            self.watchees = []
            self.filepath = None
            self.codepath = None
        return True

    def watchee_paths(self):
        return [w[1] for w in self.watchees]
    
    def renderable_error(self):
        stack = traceback.format_exc()
        print(stack)
        r = Rect(1200, 300)
        render = renderable(r)
        res = DATPens([
            DATPen().rect(r).f(coldtype.Gradient.V(r,
            coldtype.hsl(_random.random(), l=0.3),
            coldtype.hsl(_random.random(), l=0.3))),
        ])
        render.show_error = stack.split("\n")[-2]
        return render, res

    def show_error(self):
        if self.playing > 0:
            self.playing = -1
        print("============================")
        print(">>> Error in source file <<<")
        print("============================")
        render, res = self.renderable_error()
        self.previews_waiting_to_paint.append([render, res, None])
    
    def show_message(self, message, scale=1):
        print(message)
    
    def apply_syntax_mods(self, source_code):
        if self.disable_syntax_mods:
            return source_code
        
        self._codepath_offset = 0
        
        def inline_other(x):
            #cwd = self.filepath.relative_to(Path.cwd())
            cwd = Path.cwd()
            #print(">>>>>>>>>>>>>>>>>>>>>>>", cwd)
            path = Path(cwd / (x.group(1).replace(".", "/")+".py"))
            if path not in self.watchee_paths():
                self.watchees.append([Watchable.Source, path, None])
            src = path.read_text()
            self._codepath_offset = len(src.split("\n"))
            return src

        source_code = re.sub(r"from ([^\s]+) import \* \#INLINE", inline_other, source_code)
        source_code = re.sub(r"\-\.[A-Za-z_ƒ]+([A-Za-z_0-9]+)?\(", ".nerp(", source_code)
        source_code = re.sub(r"([\s]+)Ƨ\(", r"\1nerp(", source_code)
        source_code = re.sub(r"λ\s?([/\.\@]{1,2})", r"lambda xxx: xxx\1", source_code)
        #source_code = re.sub(r"λ\.", "lambda x: x.", source_code)
        source_code = re.sub(r"λ", "lambda ", source_code)
        #source_code = re.sub(r"ßDPS\(([^\)]+)\)", r"(ß:=DPS(\1))", source_code)

        while "nerp(" in source_code:
            start = source_code.find("nerp(")
            end = -1
            i = 5
            depth = 1
            c = source_code[start+i]
            while depth > 0 and c:
                #print(c, depth)
                if c == "(":
                    depth += 1
                elif c== ")":
                    depth -= 1
                i += 1
                c = source_code[start+i]
            end = start+i
            source_code = source_code[:start] + "noop()" + source_code[end:]
            #print(start, end)
        
        return source_code

    def reload(self, trigger):
        if skia and SkiaPen:
            skfx.SKIA_CONTEXT = self.context

        if not self.filepath:
            self.program = dict(no_filepath=True)
            pass # ?
        else:
            if self.filepath.suffix == ".rst":
                doctree = publish_doctree(self.filepath.read_text())
                def is_code_block(node):
                    if node.tagname == "literal_block":
                        classes = node.attributes["classes"]
                        # ok the "ruby" here is an ugly hack but it's so I can hide certain code
                        # from a printed rst
                        if "code" in classes and ("python" in classes or "ruby" in classes):
                            return True
                    return False
                code_blocks = doctree.traverse(condition=is_code_block)
                source_code = [block.astext() for block in code_blocks]
                if self.codepath:
                    self.codepath.unlink()
                with tempfile.NamedTemporaryFile("w", prefix="coldtype_rst_src", suffix=".py", delete=False) as tf:
                    tf.write("\n".join(source_code))
                    self.codepath = Path(tf.name)
            
            elif self.filepath.suffix == ".py":
                if self.disable_syntax_mods:
                    self.codepath = self.filepath
                else:
                    source_code = self.apply_syntax_mods(self.filepath.read_text())
                    if self.codepath:
                        self.codepath.unlink()
                    with tempfile.NamedTemporaryFile("w", prefix="coldtype_py_mod", suffix=".py", delete=False) as tf:
                        tf.write(source_code)
                        self.codepath = Path(tf.name)
            else:
                raise Exception("No code found in file!")
            
            try:
                self.state.reset()
                self.program = run_path(str(self.codepath), init_globals={
                    "__COLDTYPE__": True,
                    "__CONTEXT__": self.context,
                    "__FILE__": self.filepath,
                    "__sibling__": partial(sibling, self.filepath)
                })

                full_restart = False

                for r in self.renderables(Action.PreviewStoryboardReload):
                    if isinstance(r, animation):
                        if r.name not in self.state._frame_offsets:
                            full_restart = True
                            for i, s in enumerate(r.storyboard):
                                self.state.add_frame_offset(r.name, s)
                        else:
                            lasts = self.state._initial_frame_offsets[r.name]
                            if str(lasts) != str(r.storyboard):
                                del self.state._frame_offsets[r.name]
                                del self.state._initial_frame_offsets[r.name]
                                for s in r.storyboard:
                                    self.state.add_frame_offset(r.name, s)
                        
                        self.last_animation = r
                        if pyaudio and not self.args.is_subprocess and r.audio:
                            self.paudio_source = soundfile.SoundFile(r.audio, "r+")
                            #self.paudio_source = wave.open(str(r.audio), "rb")
                            self.paudio_preview = 0
                        
                        if not r.audio:
                            self.paudio_source = None
                
                if full_restart:
                    fos = {}
                    if self.args.frame_offsets:
                        fos = eval(self.args.frame_offsets)
                        for k, v in fos.items():
                            self.state.adjust_keyed_frame_offsets(k, lambda i, o: v[i])
                    
                if self.program.get("COLDTYPE_NO_WATCH"):
                    return True
            except SystemExit:
                self.on_exit(restart=False)
                return True
            except Exception as e:
                self.program = None
                self.show_error()
    
    def animation(self):
        renderables = self.renderables(Action.PreviewStoryboard)
        for r in renderables:
            if isinstance(r, animation):
                return r
    
    def buildrelease_fn(self, fnname="release"):
        candidate = None
        for k, v in self.program.items():
            if k == fnname:
                candidate = v
        return candidate
    
    def normalize_fmt(self, render):
        if self.args.format:
            render.fmt = self.args.format
        if self.args.rasterizer:
            render.rasterizer = self.args.rasterizer

        if render.rasterizer == "skia" and render.fmt in ["png", "pdf"] and skia is None:
            if not self.rasterizer_warning:
                self.rasterizer_warning = True
                print(f"RENDERER> SVG (skia-python not installed)")
            render.rasterizer = "svg"
            render.fmt = "svg"
        elif render.rasterizer == "drawbot" and render.fmt in ["png", "pdf"] and db is None:
            if not self.rasterizer_warning:
                self.rasterizer_warning = True
                print(f"RENDERER> SVG (no drawbot)")
            render.rasterizer = "svg"
            render.fmt = "svg"
    
    def renderables(self, trigger):
        _rs = []
        for k, v in self.program.items():
            if isinstance(v, renderable) and not v.hidden:
                if v not in _rs:
                    self.normalize_fmt(v)
                    _rs.append(v)
        
        for r in _rs:
            output_folder = self.render_to_output_folder(r)
            r.output_folder = output_folder

        if any([r.solo for r in _rs]):
            _rs = [r for r in _rs if r.solo]
        
        if len(self.viewer_solos) > 0:
            self.viewer_solos = [vs%len(_rs) for vs in self.viewer_solos]

            solos = []
            for i, r in enumerate(_rs):
                if i in self.viewer_solos:
                    solos.append(r)
            _rs = solos
        
        for _r in _rs:
            caps = _r.cv2caps
            if caps is not None:
                import cv2
                for cap in caps:
                    if cap not in self.state.cv2caps:
                        self.state.cv2caps[cap] = cv2.VideoCapture(cap)
            
        if self.function_filters:
            function_patterns = self.function_filters
            print(">>>", function_patterns)
            matches = []
            for r in _rs:
                for fp in function_patterns:
                    try:
                        if re.match(fp, r.name) and r not in matches:
                            matches.append(r)
                    except re.error as e:
                        print("ff regex compilation error", e)
            if len(matches) > 0:
                _rs = matches
            else:
                print(">>> no matches for ff")
        
        if self.args.monitor_lines and trigger != Action.RenderAll:
            func_name = file_and_line_to_def(self.codepath, self.line_number)
            matches = [r for r in _rs if r.name == func_name]
            print(">", matches)
            if len(matches) > 0:
                return matches

        if len(_rs) == 0:
            r = renderable((500, 500))
            def draw(r):
                #ct = coldtype.StyledString("CT", coldtype.Style("assets/ColdtypeObviously-VF.ttf", 500, wdth=0)).pen().round(1)
                #print(ct.value)
                ct = DATPen().vl([('moveTo', [(70.0, 236.0)]), ('qCurveTo', [(76.5, 236.0), (83.0, 237.5), (86.0, 238.0)]), ('qCurveTo', [(80.5, 214.0), (66.0, 153.5), (50.0, 89.0), (35.5, 28.0), (30.0, 3.5)]), ('qCurveTo', [(27.5, 2.5), (20.5, -0.5), (13.0, -2.0), (8.5, -2.0)]), ('qCurveTo', [(-2.0, -2.0), (-15.5, 6.0), (-20.0, 29.5), (-14.5, 74.5), (1.5, 148.5), (15.5, 203.5)]), ('qCurveTo', [(29.5, 259.0), (49.5, 328.0), (67.0, 364.0), (85.5, 377.0), (97.5, 377.0)]), ('qCurveTo', [(106.0, 377.0), (118.0, 374.5), (121.5, 372.0)]), ('qCurveTo', [(116.0, 352.0), (108.0, 323.0), (101.5, 297.5), (94.0, 268.5), (88.5, 247.5)]), ('qCurveTo', [(85.0, 248.5), (75.5, 250.0), (70.5, 250.0)]), ('qCurveTo', [(66.5, 250.0), (60.5, 247.5), (59.5, 244.0)]), ('qCurveTo', [(59.0, 240.0), (64.5, 236.0), (70.0, 236.0)]), ('closePath', []), ('moveTo', [(119.5, 289.5)]), ('lineTo', [(168.0, 289.5)]), ('qCurveTo', [(156.5, 242.5), (133.0, 148.5), (113.5, 67.5), (99.0, 10.5), (96.5, 0.0)]), ('qCurveTo', [(91.5, 0.5), (77.0, 1.0), (71.5, 1.0)]), ('qCurveTo', [(65.5, 1.0), (51.5, 0.5), (46.0, 0.0)]), ('qCurveTo', [(49.0, 10.5), (64.0, 67.5), (84.5, 148.5), (108.0, 242.5), (119.5, 289.5)]), ('closePath', []), ('moveTo', [(127.5, 375.0)]), ('lineTo', [(202.0, 375.0)]), ('qCurveTo', [(200.0, 365.5), (193.0, 337.5), (189.0, 323.5)]), ('qCurveTo', [(186.0, 310.5), (178.5, 281.0), (176.0, 270.0)]), ('qCurveTo', [(166.0, 270.0), (147.0, 270.5), (139.0, 270.5)]), ('qCurveTo', [(131.5, 270.5), (112.0, 270.0), (101.5, 270.0)]), ('qCurveTo', [(104.5, 281.0), (112.0, 310.5), (115.0, 323.5)]), ('qCurveTo', [(119.0, 337.5), (125.5, 365.5), (127.5, 375.0)]), ('closePath', [])])
                return DATPens([
                    DATPen().rect(r).f(coldtype.Gradient.Vertical(r,
                    coldtype.hsl(_random.random()),
                    coldtype.hsl(_random.random()))),
                    ct.align(r.inset(20), "mnx", "mxy").f(1, 0.25)
                    ])
            r.__call__(draw)
            r.blank_renderable = True
            print(">>> No renderables found <<<")
            _rs.append(r)

        return _rs
    
    def render_to_output_folder(self, render):
        if self.args.output_folder:
            return Path(self.args.output_folder).expanduser().resolve()
        elif render.dst:
            return render.dst / (render.custom_folder or render.folder(self.filepath))
        else:
            return (self.filepath.parent if self.filepath else Path(os.getcwd())) / "renders" / (render.custom_folder or render.folder(self.filepath))
    
    def add_paths_to_passes(self, trigger, render, indices):
        output_folder = self.render_to_output_folder(render)
        
        if render.prefix is None:
            if self.filepath is not None:
                prefix = f"{self.filepath.stem}_"
            else:
                prefix = None
        else:
            prefix = render.prefix

        fmt = render.fmt
        rps = []
        for rp in render.passes(trigger, self.state, indices):
            output_path = output_folder / f"{prefix}{rp.suffix}.{fmt}"
            rp.output_path = output_path
            rp.action = trigger
            rps.append(rp)
        
        return output_folder, prefix, fmt, rps

    
    def _single_thread_render(self, trigger, indices=[]) -> Tuple[int, int]:
        if not self.args.is_subprocess:
            start = ptime.time()
        
        if len(self.previews_waiting_to_paint) > 0:
            return 0, 0, []

        prev_renders = self.last_renders
        renders = self.renderables(trigger)
        self.last_renders = renders
        preview_count = 0
        render_count = 0
        output_folder = None
        try:
            for render in renders:
                render.codepath = self.codepath
                render.filepath = self.filepath

                for watch, flag in render.watch:
                    if isinstance(watch, Font) and not watch.cacheable:
                        if watch.path not in self.watchee_paths():
                            self.watchees.append([Watchable.Font, watch.path, flag])
                        for ext in watch.font.getExternalFiles():
                            if ext not in self.watchee_paths():
                                self.watchees.append([Watchable.Font, ext, flag])
                    elif watch not in self.watchee_paths():
                        self.watchees.append([Watchable.Generic, watch, flag])
                
                did_render = False
                output_folder, prefix, fmt, passes = self.add_paths_to_passes(trigger, render, indices)
                render.last_passes = passes

                #if trigger == Action.RenderAll:
                #    shutil.rmtree(output_folder, ignore_errors=True)

                previewing = (trigger in [
                    Action.Initial,
                    Action.Resave,
                    Action.PreviewStoryboard,
                    Action.PreviewIndices,
                    Action.PreviewStoryboardReload,
                ])
                
                rendering = (self.args.save_renders or trigger in [
                    Action.RenderAll,
                    Action.RenderWorkarea,
                    Action.RenderIndices,
                ])

                self.state.previewing = previewing # TODO too janky?
                self.state.preview_scale = self.preview_scale()
                
                for rp in passes:
                    output_path = rp.output_path

                    try:
                        if render.direct_draw:
                            result = None
                        else:
                            # repopulate last_result across a save
                            if not render.last_result:
                                if len(prev_renders) > 0:
                                    for pr in prev_renders:
                                        if pr.name == render.name and pr.last_result:
                                            render.last_result = pr.last_result
                            result = render.normalize_result(render.run(rp, self.state))
                        
                        if self.state.request:
                            self.requests_waiting.append([render, str(self.state.request), None])
                            self.state.request = None
                        
                        if self.state.callback:
                            self.requests_waiting.append([render, self.state.callback, "callback"])
                            self.state.callback = None

                        if not result and not render.direct_draw:
                            #print(">>> No result")
                            result = DATPen().rect(render.rect).f(None)

                        if previewing:
                            if render.direct_draw:
                                self.previews_waiting_to_paint.append([render, None, rp])
                            else:
                                preview_result = render.normalize_result(render.runpost(result, rp, self.state))
                                preview_count += 1
                                if preview_result:
                                    self.previews_waiting_to_paint.append([render, preview_result, rp])
                        
                        if rendering:
                            did_render = True
                            if False:
                                pass
                            else:
                                if render.preview_only:
                                    continue
                                render_count += 1
                                output_path.parent.mkdir(exist_ok=True, parents=True)
                                if render.self_rasterizing:
                                    print(">>> self-rasterized...", output_path.relative_to(Path.cwd()))
                                else:
                                    if render.direct_draw:
                                        self.rasterize(partial(render.run, rp, self.state), render, output_path)
                                    else:
                                        self.rasterize(result or DATPen(), render, output_path)
                                    # TODO a progress bar?
                                    print(">>> saved...", str(output_path.relative_to(Path.cwd())))
                    except:
                        self.show_error()
        except:
            self.show_error()
        if render_count > 0:
            self.show_message(f"Rendered {render_count}")
        
        if not self.args.is_subprocess and self.args.show_time:
            print("TIME >>>", ptime.time() - start)
        
        return preview_count, render_count, renders

    def render(self, trigger, indices=[]) -> Tuple[int, int]:
        if self.args.is_subprocess: # is a child process of a multiplexed render
            if trigger != Action.RenderIndices:
                raise Exception("Invalid child process render action", trigger)
                return 0, 0
            else:
                p, r, _ = self._single_thread_render(trigger, indices=indices)
                if not self.args.no_sound:
                    os.system("afplay /System/Library/Sounds/Pop.aiff")
                self.exit_code = 5 # mark as child-process
                return p, r
        
        elif self.multiplexing and self.animation():
            if trigger in [Action.RenderAll, Action.RenderWorkarea]:
                all_frames = self.animation().all_frames()
                if trigger == Action.RenderAll:
                    frames = all_frames
                elif trigger == Action.RenderWorkarea:
                    timeline = self.animation().timeline
                    try:
                        frames = list(timeline.workareas[0])
                    except:
                        frames = all_frames
                    if len(frames) == 0:
                        frames = all_frames
                self.render_multiplexed(frames)
                trigger = Action.RenderIndices
                indices = [0, all_frames[-1]] # always render first & last from main, to trigger a filesystem-change detection in premiere

        elif self.animation() and trigger == Action.RenderWorkarea:
            all_frames = self.animation().all_frames()
            self._single_thread_render(Action.RenderIndices, [0, all_frames[-1]])
        
        preview_count, render_count, renders = self._single_thread_render(trigger, indices)
        
        if not self.args.is_subprocess and render_count > 0:
            for render in renders:
                result = render.package(self.filepath, self.render_to_output_folder(render))
                if result:
                    self.previews_waiting_to_paint.append([render, result, None])
                else:
                    self.action_waiting = Action.PreviewStoryboard

            self.send_to_external(None, rendered=True)

        return preview_count, render_count
    
    def render_multiplexed(self, frames):
        start = ptime.time()

        print("TC", self.args.thread_count)
        
        group = math.floor(len(frames) / self.args.thread_count)
        ordered_frames = list(frames) #list(range(frames[0], frames[0]+len(frames)))
        shuffle(ordered_frames)
        #subslices = list(chunks(ordered_frames, group))
        subslices = [list(s) for s in distribute(self.args.thread_count, ordered_frames)]

        print(subslices)
        
        self.reset_renderers()
        self.running_renderers = []
        self.completed_renderers = []

        #logfile = filepath.parent.joinpath(f"{filepath.stem}-log.txt")
        #log = open(logfile, "w")

        for subslice in subslices:
            print("slice >", len(subslice))
            if len(subslice) == 0:
                continue
            sargs = [
                "coldtype",
                sys.argv[1],
                "-i", ",".join([str(s) for s in subslice]),
                "-isp",
                "-s", str(self.args.scale),
            ]
            r = self.args.rasterizer
            if r:
                sargs.append("-r", r)
            if self.args.no_sound:
                sargs.append("-ns")
            if self.args.cpu_render or skia is None:
                sargs.append("-cpu")
            #print(sys.argv)
            #print(sargs)
            #return
            renderer = Popen(sargs) #stdout=log)
            self.running_renderers.append(renderer)
        
        while self.running_renderers.count(None) != len(self.running_renderers):
            for idx, renderer in enumerate(self.running_renderers):
                if renderer:
                    retcode = renderer.poll()
                    if retcode == 5:
                        self.running_renderers[idx] = None
            ptime.sleep(.1)

        print("TIME >>>", ptime.time() - start)
        if not self.args.no_sound:
            os.system("afplay /System/Library/Sounds/Frog.aiff")
    
    def rasterize(self, content, render, path):
        if render.self_rasterizing:
            print("Self rasterizing")
            return
        
        scale = int(self.args.scale)
        rasterizer = self.args.rasterizer or render.rasterizer

        if rasterizer == "drawbot":
            from coldtype.pens.rendererdrawbotpen import RendererDrawBotPen
            RendererDrawBotPen.Composite(content, render.rect, str(path), scale=scale)
        elif rasterizer == "skia":
            if not skia:
                raise Exception("pip install skia-python")
            if render.fmt == "png":
                content = content.ch(skfx.precompose(render.rect))
                render.last_result = content
                if render.bg_render:
                    content = DATPens([
                        DATPen().rect(render.rect).f(render.bg),
                        content
                    ])
                SkiaPen.Composite(content, render.rect, str(path), scale=scale, context=None if self.args.cpu_render else self.context, style=render.style)
            elif render.fmt == "pdf":
                SkiaPen.PDFOnePage(content, render.rect, str(path), scale=scale)
            elif render.fmt == "svg":
                SkiaPen.SVG(content, render.rect, str(path), scale=scale)
            else:
                print("> Skia render not supported for ", render.fmt)
        elif rasterizer == "svg":
            path.write_text(SVGPen.Composite(content, render.rect, viewBox=render.viewBox))
        elif rasterizer == "pickle":
            pickle.dump(content, open(path, "wb"))
        else:
            raise Exception(f"rasterizer ({rasterizer}) not supported")
    
    def reload_and_render(self, trigger, watchable=None, indices=None):
        #self.playing = 0

        if self.args.is_subprocess and not self.args.cpu_render:
            if not glfw.init():
                raise RuntimeError('glfw.init() failed')
            glfw.window_hint(glfw.VISIBLE, glfw.FALSE)
            glfw.window_hint(glfw.STENCIL_BITS, 8)
            window = glfw.create_window(640, 480, '', None, None)
            glfw.make_context_current(window)
            self.context = skia.GrDirectContext.MakeGL()

        wl = len(self.watchees)
        self.window_scrolly = 0

        try:
            should_halt = self.reload(trigger)
            if should_halt:
                return True
            if self.program:
                preview_count, render_count = self.render(trigger, indices=indices)
                if self.args.show_render_count:
                    print("render>", preview_count, "/", render_count)
                if self.playing < 0:
                    self.playing = 1
            else:
                print(">>>>>>>>>>>> No program loaded! <<<<<<<<<<<<<<")
        except:
            self.show_error()

        if wl < len(self.watchees) and len(self.observers) > 0:
            pprint(self.watchees)
            self.stop_watching_file_changes()
            self.watch_file_changes()

    def main(self):
        if self.dead:
            return

        if self.args.memory:
            tracemalloc.start(10)
            self._last_memory = -1
        try:
            self.start()
        except KeyboardInterrupt:
            #print("INTERRUPT")
            self.on_exit()
        if self.args.show_exit_code:
            print("exit>", self.exit_code)
        sys.exit(self.exit_code)
    
    def set_title(self, text):
        if glfw and not self.args.no_viewer:
            glfw.set_window_title(self.window, text)
    
    def get_content_scale(self):
        u_scale = self.py_config.get("WINDOW_SCALE")
        
        if self.args.window_content_scale == None:
            if u_scale:
                return u_scale
            elif glfw and not self.args.no_viewer:
                if self.primary_monitor:
                    return glfw.get_monitor_content_scale(self.primary_monitor)[0]
                else:
                    return glfw.get_window_content_scale(self.window)[0]
            else:
                return 1
        else:
            return self.args.window_content_scale
    
    def initialize_gui_and_server(self):
        if self.args.websocket_external:
            self.state.external_url = f"ws://{self.args.websocket_external}"
            print(self.state.external_url)

        if self.args.websocket or self.args.webviewer:
            global WEBSOCKET_PORT
            WEBSOCKET_PORT = self.args.port
            try:
                print("WEBSOCKET>", f"localhost:{WEBSOCKET_PORT}")
                self.server = echo_server(WEBSOCKET_PORT)
                daemon = threading.Thread(name="daemon_websocket", target=self.server.serveforever)
                daemon.setDaemon(True)
                daemon.start()
            except OSError:
                self.server = None
        else:
            self.server = None
        
        if self.args.webviewer:
            port = self.args.webviewer_port
            if port != 0:
                print("WEBVIEWER>", f"localhost:{port}")

                def start_server(port):
                    httpd = HTTPServer(('', port), WebViewerHandler)
                    httpd.serve_forever()

                daemon = threading.Thread(name='daemon_server',
                    target=start_server, args=(port,))
                daemon.setDaemon(True)
                daemon.start()
        elif not glfw and not skia:
            print("\n\n>>> To view a coldtype program, you either need to (1) install with the [viewer] optional package or (2) install with the [webviewer] optional package and run with the --webviewer flag (aka -wv)\n\n")

        if glfw and not self.args.no_viewer:
            if not glfw.init():
                raise RuntimeError('glfw.init() failed')
            glfw.window_hint(glfw.STENCIL_BITS, 8)
            
            self.transparent = self.py_config.get("WINDOW_TRANSPARENT", self.args.window_transparent)
            if self.transparent:
                glfw.window_hint(glfw.TRANSPARENT_FRAMEBUFFER, glfw.TRUE)
                glfw.window_hint(glfw.DECORATED, glfw.FALSE)
            else:
                glfw.window_hint(glfw.TRANSPARENT_FRAMEBUFFER, glfw.FALSE)
            
            self.passthrough = self.py_config.get("WINDOW_PASSTHROUGH", self.args.window_passthrough)
            if self.passthrough:
                try:
                    glfw.window_hint(0x0002000D, glfw.TRUE)
                    #glfw.window_hint(glfw.MOUSE_PASSTHROUGH, glfw.TRUE)
                except glfw.GLFWError:
                    print("failed to hint window for mouse-passthrough")

            if self.py_config.get("WINDOW_BACKGROUND", self.args.window_background):
                glfw.window_hint(glfw.FOCUSED, glfw.FALSE)
            
            if self.py_config.get("WINDOW_FLOAT", self.args.window_float):
                glfw.window_hint(glfw.FLOATING, glfw.TRUE)

            self.window = glfw.create_window(int(50), int(50), '', None, None)
            self.window_scrolly = 0

            recp = sibling(__file__, "../../assets/RecMono-CasualItalic.ttf")
            self.typeface = skia.Typeface.MakeFromFile(str(recp))
            #print(self.typeface.serialize().bytes().decode("utf-8"))
            
            o = self.py_config.get("WINDOW_OPACITY", 1)
            if self.args.window_opacity is not None:
                o = self.args.window_opacity
            glfw.set_window_opacity(self.window, max(0.1, min(1, o)))

            self._prev_scale = self.get_content_scale()
            
            glfw.make_context_current(self.window)
            glfw.set_key_callback(self.window, self.on_key)
            glfw.set_char_callback(self.window, self.on_character)
            glfw.set_mouse_button_callback(self.window, self.on_mouse_button)
            glfw.set_cursor_pos_callback(self.window, self.on_mouse_move)
            glfw.set_scroll_callback(self.window, self.on_scroll)

            #if primary_monitor:
            #    glfw.set_window_monitor(self.window, primary_monitor, 0, 0, int(50), int(50))

        try:
            midiin = rtmidi.RtMidiIn()
            lookup = {}
            self.midis = []
            for p in range(midiin.getPortCount()):
                lookup[midiin.getPortName(p)] = p

            for device, mapping in self.midi_mapping.items():
                if device in lookup:
                    mapping["port"] = lookup[device]
                    mi = rtmidi.RtMidiIn()
                    mi.openPort(lookup[device])
                    self.midis.append([device, mi])
                else:
                    if self.args.midi_info:
                        print(f">>> no midi port found with that name ({device}) <<<")
        except:
            self.midis = []
        
        if skia and glfw and not self.args.no_viewer:
            self.context = skia.GrDirectContext.MakeGL()
        else:
            self.context = None

        #self.watch_file_changes()

        if len(self.watchees) > 0:
            self.set_title(self.watchees[0][1].name)
        else:
            self.set_title("coldtype")

        self.hotkeys = None
        try:
            if self.hotkey_mapping:
                from pynput import keyboard
                mapping = {}
                for k, v in self.hotkey_mapping.items():
                    mapping[k] = partial(self.on_hotkey, k, v)
                #self.hotkeys = keyboard.GlobalHotKeys({
                #    "<cmd>+<f8>": self.on_hotkey
                #})
                self.hotkeys = keyboard.GlobalHotKeys(mapping)
                self.hotkeys.start()
        except:
            pass
    
        if pyaudio:
            self.paudio = pyaudio.PyAudio()
            self.paudio_source:soundfile.SoundFile = None
            self.paudio_stream = None
            self.paudio_rate = 0
    
        if self.args.sidecar:
            if self.args.sidecar == ".":
                self.args.sidecar = self.filepath
            dst = DataSourceThread(Path(self.args.sidecar).expanduser().resolve())
            dst.start()
            self.sidecar_threads.append(dst)
        
        if self.args.tails:
            ts = self.args.tails.split(",")
            for t in ts:
                if t in ["rf", "robofont"]:
                    t = "~/Library/Application Support/RoboFont/robofont-3-py3.log"
                tp = Path(t.strip()).expanduser().absolute()
                proc = Popen(["tail", "-f", str(tp)], stdout=PIPE, stdin=PIPE, stderr=STDOUT)
                t = threading.Thread(target=show_tail, args=(proc,), daemon=True)
                t.start()
                self.tails.append(t)

    def start(self):
        if self.args.midi_info:
            try:
                midiin = rtmidi.RtMidiIn()
                ports = range(midiin.getPortCount())
                for p in ports:
                    print(p, midiin.getPortName(p))
            except:
                print("Please run `pip install rtmidi` in your venv")
                self.on_exit()
                return
            if self.args.no_watch:
                should_halt = True
            else:
                should_halt = self.before_start()
        else:
            should_halt = self.before_start()
        
        if not self.args.no_watch:
            self.initialize_gui_and_server()
        else:
            self.window = None
            self.server = None
        
        if should_halt:
            self.on_exit()
        else:
            if self.args.all:
                self.reload_and_render(Action.RenderAll)
                if self.args.build:
                    self.on_release(build=1)
                if self.args.release:
                    self.on_release()
            elif self.args.build:
                self.reload_and_render(Action.RenderIndices, indices=[0])
                self.on_release(build=1)
                if self.args.release:
                    self.on_release()
            elif self.args.release:
                self.reload_and_render(Action.RenderIndices, indices=[0])
                self.on_release()
            elif self.args.indices:
                indices = [int(x.strip()) for x in self.args.indices.split(",")]
                self.reload_and_render(Action.RenderIndices, indices=indices)
            else:
                should_halt = self.reload_and_render(Action.Initial)
                if should_halt:
                    self.on_exit()
                    return
            self.on_start()
            if not self.args.no_watch:
                if glfw and not self.args.no_viewer:
                    self.listen_to_glfw()
                elif self.args.webviewer:
                    while True:
                        self.turn_over()
                        sleep(0.25)
            else:
                self.on_exit()
    
    def before_start(self):
        pass

    def on_start(self):
        pass

    def on_request_from_render(self, render, request, action=None):
        print("request (noop)>", render, request, action)

    def on_hotkey(self, key_combo, action):
        self.hotkey_waiting = (action, key_combo)
    
    def on_message(self, message, action):
        if action:
            enum_action = self.lookup_action(action)
            if enum_action:
                print(">", enum_action)
                self.action_waiting = enum_action
                #self.on_action(enum_action, message)
            else:
                print(">>> (", action, ") is not a recognized action")
    
    def jump_to_fn(self, fn_name):
        if self.last_animation:
            fi = self.last_animation.fn_to_frame(fn_name)
            if fi is None:
                print("fn_to_frame: no match")
                return False
            #self.last_animation.storyboard = [fi]
            self.state.adjust_all_frame_offsets(0, absolute=True)
            self.state.adjust_keyed_frame_offsets(
                self.last_animation.name, lambda i, o: fi)
            self.action_waiting = Action.PreviewStoryboard
            return True
    
    def on_remote_command(self, cmd, context):
        #print(">>>>>", cmd, context)
        try:
            kbs = KeyboardShortcut(cmd)
            if kbs:
                self.on_shortcut(kbs)
                return
        except ValueError:
            pass

        if context:
            if cmd == "show_function":
                lines = self.codepath.read_text().split("\n")
                lidx = int(context)
                lidx += self._codepath_offset
                line = lines[lidx]
                while not line.startswith("@"):
                    lidx -= 1
                    line = lines[lidx]
                fn_name = lines[lidx+1].strip("def ").split("(")[0].strip()
                self.jump_to_fn(fn_name)
        else:
            if cmd == "∑":
                self.state.exit_keylayer()
                self.action_waiting = Action.PreviewStoryboard
                return
            if self.state.keylayer == Keylayer.Cmd:
                self.state.keybuffer = cmd
                self.action_waiting = Action.PreviewStoryboard
                return

            mods = [0, 0, 0, 0]
            char = None
            i = 0
            while i < len(cmd):
                c = cmd[i]
                if c == "∆": # shift
                    mods[2] = 1
                elif c == "˚": # control
                    mods[3] = 1
                elif c == "¬": # alt
                    mods[1] = 1
                elif c == "…": # super
                    mods[0] = 1
                elif c == "∑":
                    char = glfw.KEY_ESCAPE
                else:
                    char = ord(c.upper())
                i += 1
            
            #print(">", cmd, context, char, mods)
            if char:
                self.on_potential_shortcut(char, glfw.PRESS, mods)

    def lookup_action(self, action):
        try:
            return Action(action)
        except ValueError:
            try:
                return EditAction(action)
            except ValueError:
                return None
    
    def additional_actions(self):
        return []
    
    def on_scroll(self, win, xoff, yoff):
        self.window_scrolly += yoff
        #self.on_action(Action.PreviewStoryboard)
        #print(xoff, yoff)
        #pass # TODO!
    
    def on_release(self, build=False):
        fnname = "build" if build else "release"
        release_fn = self.buildrelease_fn(fnname)
        if not release_fn:
            print(f"No `{fnname}` fn defined in source")
            return
        trigger = Action.RenderAll
        renders = self.renderables(trigger)
        output_folder = None
        all_passes = []
        try:
            for render in renders:
                if not render.preview_only:
                    output_folder, prefix, fmt, passes = self.add_paths_to_passes(trigger, render, [0])
                    for rp in passes:
                        all_passes.append(rp)

            release_fn(all_passes)
        except Exception as e:
            stack = traceback.format_exc()
            print(stack)
            print("Release failed", str(e))
    
    def allow_mouse(self):
        return True
        #return self.state.keylayer == Keylayer.Editing
    
    def on_mouse_button(self, _, btn, action, mods):
        if not self.allow_mouse():
            return
        
        pos = Point(glfw.get_cursor_pos(self.window)).scale(2) # TODO should this be preview-scale?
        pos[1] = self.last_rect.h - pos[1]
        requested_action = self.state.on_mouse_button(pos, btn, action, mods)
        if requested_action:
            self.action_waiting = requested_action
    
    def on_mouse_move(self, _, xpos, ypos):
        if not self.allow_mouse():
            return
        
        pos = Point((xpos, ypos)).scale(2)
        pos[1] = self.last_rect.h - pos[1]
        requested_action = self.state.on_mouse_move(pos)
        if requested_action:
            self.action_waiting = requested_action

    def on_character(self, _, codepoint):
        if self.state.keylayer != Keylayer.Default:
            if self.state.keylayer_shifting:
                self.state.keylayer_shifting = False
                self.on_action(Action.PreviewStoryboard)
                return
            requested_action = self.state.on_character(codepoint)
            if requested_action:
                self.action_waiting = requested_action
            
    def shortcuts(self):
        if not glfw:
            return {}
        keyed = {}
        for k, v in SHORTCUTS.items():
            modded = []
            for mods, symbol in v:
                #print([symbol_to_glfw(s) for s in mods])
                modded.append([[symbol_to_glfw(s) for s in mods], symbol_to_glfw(symbol)])
            keyed[k] = modded
        return keyed
    
    def repeatable_shortcuts(self):
        return REPEATABLE_SHORTCUTS
    
    def shortcut_to_action(self, shortcut):
        if shortcut in self.repeatable_shortcuts():
            self.paudio_preview = 1
        
        if shortcut == KeyboardShortcut.PreviewPrevMany:
            return Action.PreviewStoryboardPrevMany
        elif shortcut == KeyboardShortcut.PreviewPrev:
            return Action.PreviewStoryboardPrev
        elif shortcut == KeyboardShortcut.PreviewNextMany:
            return Action.PreviewStoryboardNextMany
        elif shortcut == KeyboardShortcut.PreviewNext:
            return Action.PreviewStoryboardNext
        
        elif shortcut == KeyboardShortcut.JumpHome:
            self.state.adjust_all_frame_offsets(0, absolute=True)
        elif shortcut == KeyboardShortcut.JumpEnd:
            self.state.adjust_all_frame_offsets(-1, absolute=True)
        
        elif shortcut == KeyboardShortcut.JumpPrev:
            self.state.adjust_keyed_frame_offsets(
                self.last_animation.name,
                lambda i, o: self.last_animation.jump(o, -1))
        elif shortcut == KeyboardShortcut.JumpNext:
            self.state.adjust_keyed_frame_offsets(
                self.last_animation.name,
                lambda i, o: self.last_animation.jump(o, +1))
        elif shortcut == KeyboardShortcut.JumpStoryboard:
            self.state.adjust_keyed_frame_offsets(
                self.last_animation.name,
                lambda i, o: self.last_animation.storyboard[i])

        elif shortcut == KeyboardShortcut.ClearLastRender:
            return Action.ClearLastRender
        elif shortcut == KeyboardShortcut.ClearRenderedFrames:
            return Action.ClearRenderedFrames
        
        elif shortcut == KeyboardShortcut.PlayRendered:
            self.paudio_preview = 1
            self.on_action(Action.RenderedPlay)
            return -1
        elif shortcut == KeyboardShortcut.PlayPreview:
            self.paudio_preview = 1
            self.playing_preloaded_frame = -1
            return Action.PreviewPlay
        elif shortcut == KeyboardShortcut.PlayPreviewSlow:
            if shortcut not in self.recurring_actions:
                self.recurring_actions[shortcut] = dict(
                    interval=0.5,
                    action=KeyboardShortcut.PreviewNext,
                    last=0)
            else:
                del self.recurring_actions[shortcut]
        elif shortcut == KeyboardShortcut.PlayAudioFrame:
            self.paudio_preview = 1
        
        elif shortcut == KeyboardShortcut.ReloadSource:
            return Action.PreviewStoryboardReload
        elif shortcut == KeyboardShortcut.RestartApp:
            self.on_action(Action.RestartRenderer)
            return -1
        elif shortcut == KeyboardShortcut.Kill:
            os.kill(os.getpid(), signal.SIGINT)
            os.system(f"kill {os.getpid()}")
            return -1
        elif shortcut == KeyboardShortcut.Quit:
            self.dead = True
            return -1
        
        elif shortcut == KeyboardShortcut.Release:
            self.on_action(Action.Release)
            return -1
        elif shortcut == KeyboardShortcut.Build:
            self.on_action(Action.Build)
            return -1
        elif shortcut == KeyboardShortcut.RenderAll:
            self.on_action(Action.RenderAll)
            return -1
        elif shortcut == KeyboardShortcut.RenderWorkarea:
            self.on_action(Action.RenderWorkarea)
            return -1
        elif shortcut == KeyboardShortcut.ToggleMultiplex:
            self.on_action(Action.ToggleMultiplex)
            return -1
        
        elif shortcut == KeyboardShortcut.KeylayerEditing:
            self.state.keylayer = Keylayer.Editing
            self.state.keylayer_shifting = True
            return -1
        elif shortcut == KeyboardShortcut.KeylayerCmd:
            self.state.keylayer = Keylayer.Cmd
            self.state.keylayer_shifting = True
        elif shortcut == KeyboardShortcut.KeylayerText:
            if self.last_animation:
                fi = self.last_animation._active_frames(self.state)[0]
                txt = self.last_animation.timeline.text_for_frame(fi)
                if txt:
                    self.state.keybuffer = list(txt)
            self.state.keylayer = Keylayer.Text
            self.state.keylayer_shifting = True
        
        elif shortcut == KeyboardShortcut.OverlayInfo:
            self.state.toggle_overlay(Overlay.Info)
        elif shortcut == KeyboardShortcut.OverlayTimeline:
            self.state.toggle_overlay(Overlay.Timeline)
        
        elif shortcut == KeyboardShortcut.PreviewScaleUp:
            self.state.mod_preview_scale(+0.1)
        elif shortcut == KeyboardShortcut.PreviewScaleDown:
            self.state.mod_preview_scale(-0.1)
        elif shortcut == KeyboardShortcut.PreviewScaleMin:
            self.state.mod_preview_scale(0, 0.1)
        elif shortcut == KeyboardShortcut.PreviewScaleMax:
            self.state.mod_preview_scale(0, 5)
        elif shortcut == KeyboardShortcut.PreviewScaleDefault:
            self.state.mod_preview_scale(0, 1)

        elif shortcut == KeyboardShortcut.WindowOpacityDown:
            o = glfw.get_window_opacity(self.window)
            glfw.set_window_opacity(self.window, min(1, o-0.1))
        elif shortcut == KeyboardShortcut.WindowOpacityUp:
            o = glfw.get_window_opacity(self.window)
            glfw.set_window_opacity(self.window, min(1, o+0.1))
        elif shortcut == KeyboardShortcut.WindowOpacityMin:
            glfw.set_window_opacity(self.window, 0.1)
        elif shortcut == KeyboardShortcut.WindowOpacityMax:
            glfw.set_window_opacity(self.window, 1)
        
        elif shortcut == KeyboardShortcut.MIDIControllersPersist:
            self.state.persist()
        elif shortcut == KeyboardShortcut.MIDIControllersClear:
            self.state.clear()
        elif shortcut == KeyboardShortcut.MIDIControllersReset:
            self.state.reset(ignore_current_state=True)
        
        elif shortcut == KeyboardShortcut.JumpToFrameFunctionDef:
            frame = self.last_animation._active_frames(self.state)[0]
            fn_prefix, fn_context = self.last_animation.frame_to_fn(frame)
            original_code = self.filepath.read_text().splitlines()
            found_line = -1
            for li, line in enumerate(original_code):
                if line.strip().startswith(fn_prefix):
                    found_line = li
            editor_cmd = self.py_config.get("EDITOR_COMMAND")
            if editor_cmd:
                os.system(editor_cmd + " -g " + str(self.filepath.relative_to(Path.cwd())) + ":" + str(found_line))
        
        elif shortcut == KeyboardShortcut.ViewerTakeFocus:
            glfw.focus_window(self.window)
        
        elif shortcut == KeyboardShortcut.ViewerSoloNone:
            self.viewer_solos = []
        elif shortcut == KeyboardShortcut.ViewerSoloNext:
            if len(self.viewer_solos):
                for i, solo in enumerate(self.viewer_solos):
                    self.viewer_solos[i] = solo + 1
        elif shortcut == KeyboardShortcut.ViewerSoloPrev:
            if len(self.viewer_solos):
                for i, solo in enumerate(self.viewer_solos):
                    self.viewer_solos[i] = solo - 1
        elif shortcut in [
            KeyboardShortcut.ViewerSolo1,
            KeyboardShortcut.ViewerSolo2,
            KeyboardShortcut.ViewerSolo3,
            KeyboardShortcut.ViewerSolo4,
            KeyboardShortcut.ViewerSolo5,
            KeyboardShortcut.ViewerSolo6,
            KeyboardShortcut.ViewerSolo7,
            KeyboardShortcut.ViewerSolo8,
            KeyboardShortcut.ViewerSolo9
            ]:
            self.viewer_solos = [int(str(shortcut)[-1])-1]
        elif shortcut == KeyboardShortcut.ToggleCapturing:
            self.state.capturing_previews = not self.state.capturing_previews
        elif shortcut == KeyboardShortcut.CaptureOnce:
            self.state.adjust_all_frame_offsets(0, absolute=True)
            self.state.capturing_previews = True
            self.state.capturing_previews_once = True
            self.playing = 1
    
    def on_shortcut(self, shortcut):
        #print(shortcut)
        waiting = self.shortcut_to_action(shortcut)
        if waiting:
            if waiting != -1:
                self.action_waiting = waiting
        else:
            self.action_waiting = Action.PreviewStoryboard
    
    def on_potential_shortcut(self, key, action, mods):
        for shortcut, options in self.all_shortcuts.items():
            for modifiers, skey in options:
                if key != skey:
                    continue

                if isinstance(mods, list):
                    mod_matches = mods
                else:
                    mod_matches = [0, 0, 0, 0]
                    for idx, mod in enumerate([glfw.MOD_SUPER, glfw.MOD_ALT, glfw.MOD_SHIFT, glfw.MOD_CONTROL]):
                        if mod in modifiers:
                            if mods & mod:
                                mod_matches[idx] = 1
                        elif mod not in modifiers:
                            if not (mods & mod):
                                mod_matches[idx] = 1
                
                #print(shortcut, mod_matches, all(mod_matches))
                mod_match = all(mod_matches)
                
                if not mod_match and len(modifiers) == 0 and isinstance(mods, list):
                    mod_match = True
                
                if len(modifiers) == 0 and mods != 0 and not isinstance(mods, list):
                    mod_match = False
                
                if mod_match and key == skey:
                    if (action == glfw.REPEAT and shortcut in self.repeatable_shortcuts()) or action == glfw.PRESS:
                        #print(shortcut, modifiers, skey, mod_match)
                        return self.on_shortcut(shortcut)

    def on_key(self, win, key, scan, action, mods):
        if self.state.keylayer != Keylayer.Default:
            requested_action = self.state.on_key(win, key, scan, action, mods)
            if requested_action:
                self.action_waiting = requested_action
            return
        
        self.on_potential_shortcut(key, action, mods)
        
    def preview_audio(self, frame=None):
        #if not self.args.preview_audio:
        #    return
        
        if pyaudio and self.paudio_source:
            hz = self.paudio_source.samplerate
            width = self.paudio_source.channels

            if not self.paudio_stream or hz != self.paudio_rate:
                self.paudio_rate = hz
                self.paudio_stream = self.paudio.open(
                    format=pyaudio.paFloat32,
                    channels=width,
                    rate=hz,
                    output=True)

            if frame is None:
                audio_frame = self.last_animation._active_frames(self.state)[0]
            else:
                audio_frame = frame
            chunk = int(hz / self.last_animation.timeline.fps)

            try:
                self.paudio_source.seek(chunk*audio_frame)
                data = self.paudio_source.read(chunk)
                data = data.astype(np.float32).tostring()
                self.paudio_stream.write(data)
            except wave.Error:
                print(">>> Could not read audio at frame", audio_frame)
    
    def stdin_to_action(self, stdin):
        action_abbrev, *data = stdin.split(" ")
        if action_abbrev == "ps":
            self.state.preview_scale = max(0.1, min(5, float(data[0])))
            return Action.PreviewStoryboard, None
        elif action_abbrev == "a":
            return Action.RenderAll, None
        elif action_abbrev == "w":
            return Action.RenderWorkarea, None
        elif action_abbrev == "pf":
            return Action.PreviewIndices, [int(i.strip()) for i in data]
        elif action_abbrev == "r":
            return Action.RestartRenderer, None
        elif action_abbrev == "l":
            return Action.Release, None
        elif action_abbrev == "m":
            return Action.ToggleMultiplex, None
        elif action_abbrev == "rp":
            self.reset_filepath(data[0])
            return Action.Resave, None
        elif action_abbrev == "ff":
            self.function_filters = data
            return Action.PreviewStoryboard, None
        else:
            enum_action = self.lookup_action(action_abbrev)
            if enum_action:
                return enum_action, None
            else:
                return None, None

    def on_stdin(self, stdin):
        #if self.jump_to_fn(stdin):
        #    return
        action, data = self.stdin_to_action(stdin)
        if action:
            if action == Action.PreviewIndices:
                self.render(action, indices=data)
            elif action == Action.RestartRenderer:
                self.on_exit(restart=True)
            else:
                self.on_action(action)

    def on_action(self, action, message=None) -> bool:
        #if action != Action.PreviewStoryboardNext:
        #    print("ACTION", action)

        if action in [Action.RenderAll, Action.RenderWorkarea, Action.PreviewStoryboardReload]:
            self.reload_and_render(action)
            return True
        elif action in [Action.PreviewStoryboard]:
            self.render(Action.PreviewStoryboard)
        elif action in [
            Action.PreviewStoryboardNextMany,
            Action.PreviewStoryboardPrevMany,
            Action.PreviewStoryboardNext,
            Action.PreviewStoryboardPrev,
            Action.PreviewPlay]:
            if action == Action.PreviewPlay:
                self.playing_preloaded_frame = -1
                if self.playing == 0:
                    self.playing = 1
                else:
                    self.playing = 0
            if action == Action.PreviewStoryboardPrevMany:
                self.state.adjust_all_frame_offsets(-self.many_increment)
            elif action == Action.PreviewStoryboardPrev:
                self.state.adjust_all_frame_offsets(-1)
            elif action == Action.PreviewStoryboardNextMany:
                self.state.adjust_all_frame_offsets(+self.many_increment)
            elif action == Action.PreviewStoryboardNext:
                self.state.adjust_all_frame_offsets(+1)
            self.render(Action.PreviewStoryboard)
        elif action == Action.RenderedPlay:
            self.playing = 0
            if self.playing_preloaded_frame >= 0:
                self.playing_preloaded_frame = -1
                self.preloaded_frames = []
            else:
                anm = self.animation()
                passes = self.add_paths_to_passes(Action.RenderAll, anm, anm.all_frames())[-1]
                self.preload_frames(passes)
        elif action == Action.Build:
            self.on_release(build=True)
        elif action == Action.Release:
            self.on_release()
        elif action == Action.ArbitraryCommand:
            self.on_stdin(message.get("input"))
            return True
        elif action == Action.RestartRenderer:
            self.on_exit(restart=True)
        elif action == Action.Kill:
            os.kill(os.getpid(), signal.SIGINT)
            #self.on_exit(restart=False)
        elif action == Action.ToggleMultiplex:
            self.multiplexing = not self.multiplexing
            print(">>> MULTIPLEXING?", self.multiplexing)
        elif action == Action.ClearLastRender:
            self.last_render_cleared = True
            for r in self.renderables(Action.PreviewStoryboard):
                r.last_result = None
            self.action_waiting = Action.PreviewStoryboard
        elif action == Action.ClearRenderedFrames:
            for r in self.renderables(Action.PreviewStoryboard):
                shutil.rmtree(r.output_folder, ignore_errors=True)
            print("Deleted rendered version")
        elif message.get("serialization"):
            ptime.sleep(0.5)
            self.reload(Action.Resave)
            print(">>>>>>>>>>>", self.animation().timeline.cti)
            cw = self.animation().timeline.find_workarea()
            print("WORKAREA", cw)
            if cw:
                start, end = cw
                self.send_to_external(None, workarea_update=True, start=start, end=end)
            else:
                print("No CTI/trackGroups found")
        elif action in EditAction:
            if action in [EditAction.SelectWorkarea]:
                self.send_to_external(action, serialization_request=True)
            else:
                self.send_to_external(action, edit_action=True)
        else:
            return False
    
    def send_to_external(self, action, **kwargs):
        if not self.server:
            return

        if action in ["save", "jump_to_def"]:
            for _, client in self.server.connections.items():
                client.sendMessage(json.dumps({"midi_shortcut":action, "file":str(self.filepath), **kwargs}))
            return
        
        animation = self.animation()
        if animation and animation.timeline:
            print("EVENT", action, kwargs)
            if action:
                kwargs["action"] = action.value
            kwargs["prefix"] = self.filepath.stem
            kwargs["fps"] = animation.timeline.fps
            for _, client in self.server.connections.items():
                client.sendMessage(json.dumps(kwargs))
    
    def process_ws_message(self, message):
        try:
            jdata = json.loads(message)
            if "webviewer" in jdata:
                self.action_waiting = Action.PreviewStoryboard
                return

            action = jdata.get("action")
            if action:
                self.on_message(jdata, jdata.get("action"))
            elif jdata.get("metadata") and jdata.get("path"):
                path = Path(jdata.get("path"))
                if path == self.filepath:
                    if self.args.monitor_lines:
                        self.line_number = jdata.get("line_number")
            elif jdata.get("command"):
                cmd = jdata.get("command")
                context = jdata.get("context")
                self.on_remote_command(cmd, context)
            elif jdata.get("rendered") is not None:
                idx = jdata.get("rendered")
                print("IDX>>>>>>>", idx)
                self.state.adjust_keyed_frame_offsets(
                    self.last_animation.name,
                    lambda i, o: idx)
                self.action_waiting = Action.PreviewStoryboard
                
        #except TypeError:
        #    raise TypeError("Huh")
        except:
            self.show_error()
            print("Malformed message")
    
    def listen_to_glfw(self):
        should_close = False
        should_close = glfw.window_should_close(self.window)
        while not self.dead and not should_close:
            scale_x = self.get_content_scale()
            if scale_x != self._prev_scale:
                #print(self._prev_scale, scale_x)
                self._prev_scale = scale_x
                self.on_action(Action.PreviewStoryboard)
            
            if self._should_reload:
                self._should_reload = False
                self.on_action(Action.PreviewStoryboard)
            
            t = glfw.get_time()
            td = t - self.glfw_last_time

            spf = 0.1
            if self.last_animation:
                spf = 1 / float(self.last_animation.timeline.fps)

                if td >= spf:
                    self.glfw_last_time = t
                else:
                    glfw.poll_events()
                    continue

            if self.last_animation and self.playing_preloaded_frame >= 0 and len(self.preloaded_frames) > 0:
                GL.glClear(GL.GL_COLOR_BUFFER_BIT)

                with self.surface as canvas:
                    path = self.preloaded_frames[self.playing_preloaded_frame]
                    c = self.last_animation.bg
                    canvas.clear(c.skia())
                    image = skia.Image.MakeFromEncoded(skia.Data.MakeFromFileName(str(path)))
                    canvas.drawImage(image, 0, 0)
                
                self.surface.flushAndSubmit()
                glfw.swap_buffers(self.window)

                self.preview_audio(frame=(self.playing_preloaded_frame+1)%len(self.preloaded_frames))

                self.playing_preloaded_frame += 1
                if self.playing_preloaded_frame == len(self.preloaded_frames):
                    self.playing_preloaded_frame = 0
                ptime.sleep(0.01)
            else:
                ptime.sleep(0.025)
                self.glfw_last_time = t
                self.last_previews = self.turn_over()
                global last_line
                if self.state.cmd:
                    cmd = self.state.cmd
                    self.state.reset_keystate()
                    self.on_stdin(cmd)
                elif last_line:
                    self.on_stdin(last_line.strip())
                    last_line = None
                
                #if self.playing != 0:
                    #self.action_waiting = Action.PreviewStoryboardNext
                    #self.on_action(Action.PreviewStoryboardNext)
            
            self.state.reset_keystate()
            glfw.poll_events()

            if self.state.capturing_previews and len(self.last_previews) > 0:
                if not skia or len(self.state.cv2caps) == 0:
                    print("> capturing enabled but no skia or cameras found")
                else:
                    from coldtype.capture import read_frame
                    #print(self.last_animation)
                    fo = self.state.get_frame_offsets(self.last_animation.name)

                    if self.state.capturing_previews_once:
                        fo = self.state.get_frame_offsets(self.last_animation.name)
                        #print(fo)
                        if fo[0]%self.last_animation.duration == 0:
                            self.playing = 0
                            self.state.capturing_previews = False
                            self.state.capturing_previews_once = False
                            self.state.adjust_all_frame_offsets(0, absolute=True)
                            print("/captured-once")

                    sleep(self.args.capture_previews_delay)
                    for rp in self.last_previews:
                        rf = read_frame(self.state.cv2caps[0]).align(rp.render.rect)
                        rp.output_path.parent.mkdir(exist_ok=1, parents=1)
                        if True:
                            # TODO could defer this until after a capture-all action?
                            SkiaPen.Composite(DATPens([rf]), rp.render.rect, str(rp.output_path), 1, self.context)
                            print("CAPTURE:", rp.output_path)

            should_close = glfw.window_should_close(self.window)
        self.on_exit(restart=False)
    
    def preview_scale(self):
        return self.state.preview_scale
    
    def create_surface(self, rect):
        #print("NEW SURFACE", rect)
        backend_render_target = skia.GrBackendRenderTarget(
            int(rect.w), int(rect.h), 0, 0,
            skia.GrGLFramebufferInfo(0, GL.GL_RGBA8))
        self.surface = skia.Surface.MakeFromBackendRenderTarget(
            self.context,
            backend_render_target,
            skia.kBottomLeft_GrSurfaceOrigin,
            skia.kRGBA_8888_ColorType,
            skia.ColorSpace.MakeSRGB())
        assert self.surface is not None
    
    def turn_over_webviewer(self):
        renders = []
        try:
            title = self.watchees[0][1].name
        except:
            title = "coldtype"

        for idx, (render, result, rp) in enumerate(self.previews_waiting_to_paint):
            if self.args.format == "canvas":
                renders.append(dict(fmt="canvas", jsonpen=JSONPen.Composite(result, render.rect), rect=[*render.rect], bg=[*render.bg]))
            else:
                renders.append(dict(fmt="svg", svg=SVGPen.Composite(result, render.rect, viewBox=render.viewBox), rect=[*render.rect], bg=[*render.bg]))
    
        if renders:
            for _, client in self.server.connections.items():
                if hasattr(client, "webviewer") and client.webviewer:
                    client.sendMessage(json.dumps({"renders":renders, "title":title}))
        
        return []
    
    def turn_over_glfw(self):
        dscale = self.preview_scale()
        rects = []

        render_previews = len(self.previews_waiting_to_paint) > 0
        if not render_previews:
            return []

        if render_previews:
            if pyaudio and self.paudio_source:
                if self.paudio_preview == 1:
                    self.preview_audio()
                    if self.playing == 0:
                        self.paudio_preview = 0

        w = 0
        llh = -1
        lh = -1
        h = 0
        for render, result, rp in self.previews_waiting_to_paint:
            if hasattr(render, "show_error"):
                sr = render.rect
            else:
                sr = render.rect.scale(dscale, "mnx", "mny").round()
            w = max(sr.w, w)
            if render.layer:
                rects.append(Rect(0, llh, sr.w, sr.h))
            else:
                rects.append(Rect(0, lh+1, sr.w, sr.h))
                llh = lh+1
                lh += sr.h + 1
                h += sr.h + 1
        h -= 1
        
        frect = Rect(0, 0, w, h)
        if frect != self.last_rect:
            self.create_surface(frect)

        if not self.last_rect or frect != self.last_rect:
            primary_monitor = None
            if self.args.monitor_name:
                remn = self.args.monitor_name
                monitors = glfw.get_monitors()
                matches = []
                for monitor in monitors:
                    mn = glfw.get_monitor_name(monitor)
                    if remn == "list":
                        print(">>> MONITOR >>>", mn)
                    elif remn in str(mn):
                        matches.append(monitor)
                if len(matches) > 0:
                    primary_monitor = matches[0]
                if primary_monitor:
                    self.primary_monitor = primary_monitor
                else:
                    self.primary_monitor = glfw.get_primary_monitor()

            m_scale = self.get_content_scale()
            scale_x, scale_y = m_scale, m_scale
            
            #if not DARWIN:
            #    scale_x, scale_y = 1.0, 1.0

            ww = int(w/scale_x)
            wh = int(h/scale_y)
            glfw.set_window_size(self.window, ww, wh)

            pin = self.py_config.get("WINDOW_PIN", None)
            if self.args.window_pin:
                pin = self.args.window_pin
            
            if not isinstance(pin, str):
                print("--window-pin must be compass direction")
                pin = "NE"

            if pin:
                if primary_monitor:
                    monitor = primary_monitor
                else:
                    monitor = glfw.get_primary_monitor()
                self.primary_monitor = monitor
                work_rect = Rect(glfw.get_monitor_workarea(monitor))
                wrz = work_rect.zero()
                #print(work_rect, wrz)
                edges = Edge.PairFromCompass(pin)
                pinned = wrz.take(ww, edges[0]).take(wh, edges[1]).round()
                if edges[1] == "mdy":
                    pinned = pinned.offset(0, -30)
                pinned = pinned.flip(wrz.h)
                pinned = pinned.offset(*work_rect.origin())
                if self.args.window_pin_inset:
                    x, y = [int(n) for n in self.args.window_pin_inset.split(",")]
                    pinned = pinned.offset(-x, y)
                glfw.set_window_pos(self.window, pinned.x, pinned.y)
            else:
                glfw.set_window_pos(self.window, 0, 0)

        self.last_rect = frect

        GL.glClear(GL.GL_COLOR_BUFFER_BIT)

        did_preview = []

        with self.surface as canvas:
            canvas.clear(skia.Color4f(0.3, 0.3, 0.3, 1))
            if self.transparent:
                canvas.clear(skia.Color4f(0.3, 0.3, 0.3, 0))
            
            for idx, (render, result, rp) in enumerate(self.previews_waiting_to_paint):
                rect = rects[idx].offset((w-rects[idx].w)/2, 0).round()
                try:
                    self.draw_preview(idx, dscale, canvas, rect, (render, result, rp))
                    did_preview.append(rp)
                except Exception as e:
                    stack = traceback.format_exc()
                    print(stack)
                    paint = skia.Paint(AntiAlias=True, Color=skia.ColorRED)
                    canvas.drawString(stack.split("\n")[-2], 10, 32, skia.Font(None, 36), paint)
        
            if self.state.keylayer != Keylayer.Default and not self.args.hide_keybuffer:
                self.state.draw_keylayer(canvas, self.last_rect, self.typeface)
        
        self.surface.flushAndSubmit()
        glfw.swap_buffers(self.window)
        return did_preview
    
    def turn_over(self):
        if self.dead:
            self.on_exit()
            return
        
        if self.hotkey_waiting:
            self.execute_string_as_shortcut_or_action(*self.hotkey_waiting)
            self.hotkey_waiting = None
        
        now = ptime.time()
        for k, v in self.watchee_mods.items():
            if v and (now - v) > 1:
                #print("CAUGHT ONE")
                self.action_waiting = Action.PreviewStoryboard
                self.watchee_mods[k] = None

        if self.action_waiting:
            action_in = self.action_waiting
            self.on_action(self.action_waiting)
            if action_in != self.action_waiting:
                # TODO should be recursive?
                self.on_action(self.action_waiting)
            self.action_waiting = None

        #if self.playing > 0:
        #    self.on_action(Action.PreviewStoryboardNext)
        
        if self.server:
            msgs = []
            for k, v in self.server.connections.items():
                if hasattr(v, "messages") and len(v.messages) > 0:
                    for msg in v.messages:
                        msgs.append(msg)
                        #self.process_ws_message(msg)
                    v.messages = []
            
            for msg in msgs:
                self.process_ws_message(msg)

        if not self.args.no_midi:
            self.monitor_midi()
        
        #if len(self.waiting_to_render) > 0:
        #    for action, path in self.waiting_to_render:
        #        self.reload_and_render(action, path)
        #    self.waiting_to_render = []
        
        if glfw and not self.args.no_viewer:
            did_preview = self.turn_over_glfw()
        
        if self.args.webviewer:
            did_preview = self.turn_over_webviewer()
        
        self.state.needs_display = 0
        self.previews_waiting_to_paint = []
        self.last_render_cleared = False
    
        for render, request, action in self.requests_waiting:
            if action == "callback":
                self.action_waiting = Action.PreviewStoryboard
            else:
                self.on_request_from_render(render, request, action)
            self.requests_waiting = []

        if self.playing > 0:
            self.on_action(Action.PreviewStoryboardNext)
        
        for k, ra in self.recurring_actions.items():
            interval = ra.get("interval")
            if now - ra.get("last") > interval:
                ra["last"] = now
                print("RECURRING ACTION", k)
                self.on_shortcut(ra.get("action"))
        
        return did_preview

    def draw_preview(self, idx, scale, canvas, rect, waiter):
        if isinstance(waiter[1], Path) or isinstance(waiter[1], str):
            image = skia.Image.MakeFromEncoded(skia.Data.MakeFromFileName(str(waiter[1])))
            if image:
                canvas.drawImage(image, rect.x, rect.y)
            return
        
        render, result, rp = waiter
        error_color = coldtype.rgb(1, 1, 1).skia()
        canvas.save()
        canvas.translate(0, self.window_scrolly)
        canvas.translate(rect.x, rect.y)
        
        if not self.transparent:
            canvas.drawRect(skia.Rect(0, 0, rect.w, rect.h), skia.Paint(Color=render.bg.skia()))
        
        if not hasattr(render, "show_error"):
            canvas.scale(scale, scale)
        
        if render.clip:
            canvas.clipRect(skia.Rect(0, 0, rect.w, rect.h))
        
        #canvas.clear(coldtype.bw(0, 0).skia())
        #print("BG", render.func.__name__, render.bg)
        #canvas.clear(render.bg.skia())
        
        if render.direct_draw:
            try:
                render.run(rp, self.state, canvas)
            except Exception as e:
                stack = traceback.format_exc()
                print(stack)
                render.show_error = stack.split("\n")[-2]
                error_color = coldtype.rgb(0, 0, 0).skia()
        else:
            if render.composites:
                comp = result.ch(skfx.precompose(render.rect))
                if not self.last_render_cleared:
                    render.last_result = comp
                else:
                    render.last_result = None
            else:
                comp = result
            
            #print("DRAW---\n", comp.tree())
            render.draw_preview(1.0, canvas, render.rect, comp, rp)
        
        if hasattr(render, "blank_renderable"):
            paint = skia.Paint(AntiAlias=True, Color=coldtype.hsl(0, l=1, a=0.75).skia())
            canvas.drawString(f"{coldtype.__version__}", 359, 450, skia.Font(self.typeface, 42), paint)
            canvas.drawString("Nothing found", 297, 480, skia.Font(self.typeface, 24), paint)
        
        if hasattr(render, "show_error"):
            paint = skia.Paint(AntiAlias=True, Color=error_color)
            canvas.drawString(render.show_error, 30, 70, skia.Font(self.typeface, 50), paint)
            canvas.drawString("> See process in terminal for traceback", 30, 120, skia.Font(self.typeface, 32), paint)
        
        canvas.restore()
    
    def preload_frames(self, passes):
        for rp in passes:
            self.preloaded_frames.append(rp.output_path)
        self.playing_preloaded_frame = 0
    
    def on_modified(self, event):
        path = Path(event.src_path)
        #return
        actual_path = path
        if path.parent in self.watchee_paths():
            actual_path = path
            path = path.parent
        if path in self.watchee_paths():
            if path.suffix == ".json":
                last = self.watchee_mods.get(path)
                now = ptime.time()
                self.watchee_mods[path] = now
                if last is not None:
                    diff = now - last
                    if diff < 1:
                        #print("SKIP")
                        return
                    else:
                        #print("CONTINUE")
                        pass
                try:
                    json.loads(path.read_text())
                except json.JSONDecodeError:
                    print("Error decoding watched json", path)
                    #print(path.read_text())
                    return
            
            if path.suffix == ".py":
                try:
                    for render in self.renderables(Action.Resave):
                        for wr in render.watch_restarts:
                            if str(path) == str(wr):
                                importlib.reload(coldtype)
                                importlib.reload(coldtype.pens.datpen)
                                #self.action_waiting = Action.RestartRenderer
                except AttributeError:
                    pass
            
            idx = self.watchee_paths().index(path)
            wpath, wtype, wflag = self.watchees[idx]
            if wflag == "soft":
                self.state.watch_soft_mods[actual_path] = True
                self.action_waiting = Action.PreviewStoryboard
                return

            try:
                print(f">>> resave: {Path(event.src_path).relative_to(Path.cwd())}")
            except:
                print(f">>> resave: {event.src_path}")
            
            if self.args.memory and process:
                memory = bytesto(process.memory_info().rss)
                diff = memory - self._last_memory
                self._last_memory = memory
                print(">>> pid:{:d}/new:{:04.2f}MB/total:{:4.2f}".format(os.getpid(), diff, memory))
            
            self.action_waiting = Action.PreviewStoryboardReload
            #self.waiting_to_render = [[Action.Resave, self.watchees[idx][0]]]

    def watch_file_changes(self):
        if self.args.no_watch:
            return None
        
        if not AsyncWatchdog:
            print("> must install watchdog to watch files")
            return None

        self.observers = []
        dirs = set([w[1] if w[1].is_dir() else w[1].parent for w in self.watchees])
        for d in dirs:
            o = AsyncWatchdog(str(d), on_modified=self.on_modified, recursive=True)
            o.start()
            self.observers.append(o)
        if self.filepath:
            print(f"... watching {self.filepath.name} for changes ...")
        else:
            print(f"... no file specified, showing generic window ...")
    
    def execute_string_as_shortcut_or_action(self, shortcut, key):
        try:
            ksc = KeyboardShortcut(shortcut)
            ea = None
        except ValueError:
            try:
                ksc = None
                ea = EditAction(shortcut)
            except ValueError:
                ea = None
        
        if ksc:
            self.on_shortcut(KeyboardShortcut(shortcut))
        elif ea:
            self.on_action(EditAction(shortcut), {})
        elif not ea:
            print("No shortcut/action", key, shortcut)

    
    def monitor_midi(self):
        controllers = {}
        for device, mi in self.midis:
            msg = mi.getMessage(0)
            while msg:
                if self.args.midi_info:
                    print(device, msg)
                if msg.isNoteOn(): # Maybe not forever?
                    nn = msg.getNoteNumber()
                    shortcut = self.midi_mapping[device]["note_on"].get(nn)
                    self.execute_string_as_shortcut_or_action(shortcut, nn)
                if msg.isController():
                    cn = msg.getControllerNumber()
                    cv = msg.getControllerValue()
                    shortcut = self.midi_mapping[device].get("controller", {}).get(cn)
                    if shortcut:
                        if cv in shortcut:
                            print("shortcut!", shortcut, cv)
                            self.execute_string_as_shortcut_or_action(shortcut.get(cv), cn)
                    else:
                        controllers[device + "_" + str(cn)] = cv/127
                msg = mi.getMessage(0)
        
        if len(controllers) > 0:
            nested = {}
            for k, v in controllers.items():
                device, number = k.split("_")
                if not nested.get(device):
                    nested[device] = {}
                nested[device][str(number)] = v
            
            for device, numbers in nested.items():
                self.state.controller_values[device] = {**self.state.controller_values.get(device, {}), **numbers}

            if not self.playing:
                self.action_waiting = Action.PreviewStoryboard
            #self.on_action(Action.PreviewStoryboard, {})
    
    def stop_watching_file_changes(self):
        for o in self.observers:
            o.stop()
        
    def reset_renderers(self):
        for r in self.running_renderers:
            if r:
                r.terminate()
    
    def restart(self):
        print("> RESTARTING...")
        args = sys.argv
        
        # attempt to preserve state across reload
        fo = str(self.state._frame_offsets)
        try:
            foi = args.index("-fo")
            args[foi+1] = fo
        except ValueError:
            args.append("-fo")
            args.append(fo)
        
        os.execl(sys.executable, *(["-m"]+args))

    def on_exit(self, restart=False):
        #if self.args.watch:
        #   print(f"<EXIT(restart:{restart})>")
        #if self.webviewer_thread:
        #    self.webviewer_thread.terminate()
        if glfw and not self.args.no_viewer:
            glfw.terminate()
        if self.context:
            self.context.abandonContext()
        if self.hotkeys:
            self.hotkeys.stop()
        self.reset_renderers()
        self.stop_watching_file_changes()

        if pyaudio:
            if hasattr(self, "paudio_stream") and self.paudio_stream:
                self.paudio_stream.stop_stream()
                self.paudio_stream.close()
            if hasattr(self, "paudio") and self.paudio:
                self.paudio.terminate()
            if hasattr(self, "paudio_source") and self.paudio_source:
                self.paudio_source.close()
        
        for t in self.sidecar_threads:
            t.should_run = False
        
        #for t in self.tails:
        #    t.terminate()
        
        if self.args.memory:
            snapshot = tracemalloc.take_snapshot()
            top_stats = snapshot.statistics('traceback')

            # pick the biggest memory block
            stat = top_stats[0]
            print("%s memory blocks: %.1f KiB" % (stat.count, stat.size / 1024))
            for line in stat.traceback.format():
                print(line)
        
        if restart:
            self.restart()


def main():
    pargs, parser = Renderer.Argparser()
    Renderer(parser).main()

if __name__ == "__main__":
    main()